# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import threading
from collections import OrderedDict

from mindspore import Tensor
from mindspore.common.hook_handle import HookHandle
from mindspore.ops.operations import _inner_ops as inner

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.runtime import Runtime
from msprobe.core.common.utils import ModuleQueue, ThreadSafe
from msprobe.core.dump.data_dump.scope import ModuleRangeScope, MixRangeScope, BaseScope
from msprobe.core.common.megatron_utils import wrap_megatron_step, get_micro_step, is_megatron
from msprobe.mindspore.common.const import Const as MsConst
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.common.utils import (
    is_mindtorch,
    get_cells_and_names_with_index,
    has_kwargs_in_forward_hook,
    is_graph_mode_cell_dump_allowed,
    is_backward_hook_output_a_view
)
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig
from msprobe.mindspore.dump.dump_processor.graph_mode_cell_dump import GraphModeCellDump


def get_cell_construct(construct):
    def _construct(self, *args, **kwargs):
        if hasattr(self, 'msprobe_hook'):
            setattr(self, 'msprobe_input_kwargs', kwargs)
        return construct(self, *args, **kwargs)

    return _construct


def patch_schedules_step():
    try:
        from mindspeed.mindspore.core.pipeline_parallel import schedules
        schedules.forward_step = wrap_megatron_step(schedules.forward_step)
        schedules.backward_step = wrap_megatron_step(schedules.backward_step, is_forward=False)
        logger.info_on_rank_0("Patch mindspeed.mindspore method success.")
    except ImportError:
        logger.info_on_rank_0("No mindspeed.mindspore find.")
    except Exception as e:
        logger.info_on_rank_0(f"Patch mindspeed.mindspore method failed, detail:{str(e)}")

    try:
        from megatron.core.pipeline_parallel import schedules
        schedules.forward_step = wrap_megatron_step(schedules.forward_step)
        schedules.backward_step = wrap_megatron_step(schedules.backward_step, is_forward=False)
        logger.info_on_rank_0("Patch megatron method success.")
    except ImportError:
        logger.info_on_rank_0("No megatron find.")
    except Exception as e:
        logger.info_on_rank_0(f"Patch megatron method failed, detail:{str(e)}")


class CellProcessor:
    cell_queue = ModuleQueue()
    cell_count = {}
    cell_stack = {}
    api_parent_node = {}
    module_node = {}
    cell_bw_hook_kernels = {}
    cell_backward_pre_hook = []
    cell_backward_hook = []

    def __init__(self, scope):
        self.scope = scope if isinstance(scope, (ModuleRangeScope, MixRangeScope)) else None

    @staticmethod
    def set_and_get_calls_number(cell_name):
        if cell_name not in CellProcessor.cell_count:
            CellProcessor.cell_count[cell_name] = 0
        else:
            CellProcessor.cell_count[cell_name] += 1
        return CellProcessor.cell_count[cell_name]

    @classmethod
    def reset_cell_stats(cls):
        cls.cell_queue = ModuleQueue()
        cls.cell_count = {}
        cls.cell_stack = {}
        cls.api_parent_node = {}
        cls.module_node = {}
        cls.cell_bw_hook_kernels = {}
        cls.cell_backward_pre_hook = []
        cls.cell_backward_hook = []

    def register_cell_hook(self, models, build_hook, config: DebuggerConfig):
        if not models:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   'The model cannot be None, when level is "L0" or "mix"')

        patch_schedules_step()

        is_registered = False
        model_type = Const.MODULE if is_mindtorch() else Const.CELL
        cells_with_index_in_pynative_mode, cells_with_index_in_graph_mode = get_cells_and_names_with_index(models)
        construct_name = '_call_impl' if is_mindtorch() else '_run_construct'

        for index, cells_and_names in cells_with_index_in_pynative_mode.items():
            model = models if index == "-1" else models[int(index)]
            for name, cell in cells_and_names:
                if cell == model:
                    continue

                if not has_kwargs_in_forward_hook():
                    if not hasattr(cell.__class__, 'msprobe_construct'):
                        setattr(cell.__class__, 'msprobe_construct', True)
                        if hasattr(cell.__class__, construct_name):
                            setattr(cell.__class__, construct_name,
                                    get_cell_construct(getattr(cell.__class__, construct_name)))
                setattr(cell, 'msprobe_hook', True)

                cell_index = (index + Const.SEP) if index != "-1" else ""
                prefix = f'{model_type}{Const.SEP}{cell_index}{name}{Const.SEP}{cell.__class__.__name__}{Const.SEP}'

                forward_pre_hook = self.build_cell_hook(prefix, build_hook)
                cell.register_forward_pre_hook(forward_pre_hook)

                if not is_registered:
                    logger.info("The cell hook function is successfully mounted to the model.")
                is_registered = True

        if is_graph_mode_cell_dump_allowed(config):
            cells_and_names_in_graph_mode = []
            for index, cells_and_names in cells_with_index_in_graph_mode.items():
                model = models if index == "-1" else models[int(index)]
                for name, cell, parent_cell in cells_and_names:
                    if cell == model:
                        continue
                    cell_index = (index + Const.SEP) if index != "-1" else ""
                    cells_and_names_in_graph_mode.append((f'{cell_index}{name}', cell, parent_cell))

            if cells_and_names_in_graph_mode:
                Runtime.run_mode = MsConst.PYNATIVE_GRAPH_MODE
                GraphModeCellDump(config, cells_and_names_in_graph_mode, strict=False).handle()


    def build_cell_hook(self, cell_name, build_data_hook):
        @ThreadSafe.synchronized
        def forward_pre_hook(cell, args):
            if not Runtime.is_running:
                return args

            index = CellProcessor.set_and_get_calls_number(cell_name)
            full_forward_name = f'{cell_name}{Const.FORWARD}{Const.SEP}{index}'
            full_backward_name = f'{cell_name}{Const.BACKWARD}{Const.SEP}{index}'

            self.set_construct_info_in_pre_hook(full_forward_name)

            if not hasattr(cell, 'msprobe_forward_hook'):
                if is_mindtorch():
                    cell.register_forward_hook(forward_hook, prepend=True, with_kwargs=True)
                else:
                    forward_hook_dict = getattr(cell, '_forward_hook', OrderedDict())
                    if has_kwargs_in_forward_hook():
                        forward_hook_with_kwargs_dict = getattr(cell, '_forward_hook_with_kwargs', OrderedDict())
                        handle = HookHandle(forward_hook_dict, extra_dict=forward_hook_with_kwargs_dict)
                        forward_hook_with_kwargs_dict[handle.handle_id] = True
                    else:
                        handle = HookHandle(forward_hook_dict)
                    forward_hook_dict[handle.handle_id] = forward_hook
                    forward_hook_dict.move_to_end(handle.handle_id, last=False)

                setattr(cell, 'msprobe_forward_hook', True)

            def get_backward_hook(backward_data_hook, full_backward_name):
                @ThreadSafe.synchronized
                def backward_hook_fn(cell, grad_input, grad_output):
                    new_output = backward_data_hook(cell, grad_input, grad_output)
                    self.set_construct_info_in_hook(full_backward_name)
                    cell.has_pre_hook_called = False
                    return new_output

                return backward_hook_fn

            enable_hooked = sum(
                [isinstance(ele, Tensor) and ele.dtype not in MsConst.NonDifferentiableType for ele in args]
            )
            if enable_hooked:
                backward_hook = OrderedDict()
                hook_set = build_data_hook(BaseScope.Module_Type_Module, full_forward_name)
                backward_hook[full_backward_name] = get_backward_hook(hook_set.backward_hook, full_backward_name)
                CellProcessor.cell_backward_hook.append(backward_hook)
                bw_hook = inner.CellBackwardHook(full_backward_name, cell,
                                                 self.cell_backward_hook[-1])
                bw_hook.register_backward_hook()
                CellProcessor.cell_bw_hook_kernels[full_forward_name] = bw_hook

                args = bw_hook(args) if is_backward_hook_output_a_view() else bw_hook(*args)

            return args

        @ThreadSafe.synchronized
        def forward_hook(cell, args, kwargs_or_output, output_or_kwargs=None):
            index = CellProcessor.cell_count.get(cell_name, 0)
            full_forward_name = f'{cell_name}{Const.FORWARD}{Const.SEP}{index}'
            full_backward_name = f'{cell_name}{Const.BACKWARD}{Const.SEP}{index}'

            self.set_construct_info_in_hook(full_forward_name)

            hook_set = build_data_hook(BaseScope.Module_Type_Module, full_forward_name)
            hook_result = hook_set.forward_hook(cell, args, kwargs_or_output, output_or_kwargs)
            if hook_result is not None:
                outputs = hook_result
            else:
                outputs = output_or_kwargs if has_kwargs_in_forward_hook() else kwargs_or_output

            bw_hook = CellProcessor.cell_bw_hook_kernels.get(full_forward_name)
            if bw_hook:
                if not isinstance(outputs, (Tensor, tuple)):
                    logger.warning("For backward hooks to be called,"
                                   " cell output should be a Tensor or a tuple of Tensors"
                                   f" but received {type(outputs)}")
                if is_backward_hook_output_a_view():
                    new_outputs = bw_hook(outputs)
                else:
                    if isinstance(outputs, tuple):
                        new_outputs = bw_hook(*outputs)
                    else:
                        new_outputs = bw_hook(outputs)
                    if isinstance(outputs, tuple) and len(outputs) == 1:
                        new_outputs = (new_outputs,)
                outputs = new_outputs

            def get_backward_pre_hook(full_backward_name, backward_data_hook):
                @ThreadSafe.synchronized
                def backward_pre_hook_fn(cell, grad_output):
                    cell.has_pre_hook_called = True
                    self.set_construct_info_in_pre_hook(full_backward_name)
                    if backward_data_hook:
                        backward_data_hook(cell, (), grad_output)
                        self.set_construct_info_in_hook(full_backward_name)
                        cell.has_pre_hook_called = False

                return backward_pre_hook_fn

            backward_pre_hook = OrderedDict()
            backward_data_hook = None if bw_hook else hook_set.backward_hook
            backward_pre_hook[full_backward_name] = get_backward_pre_hook(full_backward_name, backward_data_hook)
            CellProcessor.cell_backward_pre_hook.append(backward_pre_hook)
            bw_pre_hook = inner.CellBackwardHook(full_backward_name, cell,
                                                 self.cell_backward_pre_hook[-1])
            bw_pre_hook.register_backward_pre_hook()

            if is_backward_hook_output_a_view():
                result = bw_pre_hook(outputs)
            else:
                if isinstance(outputs, tuple):
                    result = bw_pre_hook(*outputs)
                else:
                    result = bw_pre_hook(outputs)
                if isinstance(outputs, tuple):
                    if len(outputs) == 1:
                        result = (result,)
                    if len(result) != len(outputs):
                        raise TypeError(
                            f"The backward pre hook return value size is {len(result)} "
                            f"not equal to output size {len(outputs)}"
                        )
            return result

        return forward_pre_hook

    def set_construct_info_in_pre_hook(self, full_name):
        tid = threading.get_ident()
        if tid not in self.cell_stack:
            CellProcessor.cell_stack[tid] = []

        if self.cell_stack[tid]:
            CellProcessor.module_node[full_name] = self.cell_stack[tid][-1] if not is_megatron() \
                else [self.cell_stack[tid][-1], get_micro_step()]
        else:
            parent_name = CellProcessor.cell_queue.find_last(full_name)
            CellProcessor.module_node[full_name] = parent_name if not is_megatron() else [parent_name, get_micro_step()]

        CellProcessor.cell_queue.add_name(full_name)
        CellProcessor.cell_stack[tid].append(full_name)
        CellProcessor.api_parent_node[tid] = full_name if not is_megatron() else [full_name, get_micro_step()]
        if self.scope:
            self.scope.begin_module(full_name)

    def set_construct_info_in_hook(self, full_name):
        tid = threading.get_ident()
        CellProcessor.cell_queue.remove_name(full_name)
        CellProcessor.api_parent_node[tid] = None if not is_megatron() else [None, get_micro_step()]
        if self.cell_stack.get(tid):
            CellProcessor.cell_stack[tid].pop()
        if self.cell_stack.get(tid):
            CellProcessor.api_parent_node[tid] = CellProcessor.cell_stack[tid][-1] if not is_megatron() \
                else [CellProcessor.cell_stack[tid][-1], get_micro_step()]
        if self.scope:
            self.scope.end_module(full_name)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
from collections import OrderedDict
from multiprocessing import Pool, cpu_count

import numpy as np

from msprobe.core.common.const import FileCheckConst
from msprobe.core.common.file_utils import FileChecker, FileOpen, check_file_or_directory_path
from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException
from msprobe.core.compare.torchair_cmp_utils import BasicDataInfo, fill_row_data, save_compare_result_to_csv
from msprobe.infer.utils.acc_cmp import parse_torchair_dump_data, set_msaccucmp_path_from_cann

GE_GRAPH_FILE_PREFIX = 'dynamo_original_graph_'
GE_DUMP_TIME_PATTERN = 'YYYYMMDDHHMMSS'
FUSION_OP_TYPE = 'AutomaticBufferFusionOp'
DUMP_FILE_FILTER_SUFIX = ['.txt', '.npy', '.bin']
MAX_TOKEN_LEN = 12


def get_rank_id_from_torchair_data(dir_name: str):
    rank_id = -1
    rank_index = dir_name.rfind('rank')
    if dir_name.startswith('worldsize') and rank_index != -1 and str.isdigit(dir_name[rank_index + 4:]):
        rank_id = int(dir_name[rank_index + 4:])
    return rank_id


def get_torchair_ge_graph_path(my_path, rank=-1):
    if not os.path.isdir(my_path):
        return None

    ge_graph_files = []
    my_path_depth = len(my_path.split(os.sep))
    timestamp_pattern = re.compile(r"(\d+)")
    for cur_path, _, file_names in os.walk(my_path):
        for file_name in file_names:
            if rank > -1 and f'rank_{rank}_' not in file_name:
                continue
            if file_name.startswith(GE_GRAPH_FILE_PREFIX) and file_name.endswith(".txt"):
                match = timestamp_pattern.search(file_name)
                if match:
                    full_path = os.path.join(cur_path, file_name)
                    timestamp = int(match.group(1))
                    ge_graph_files.append((full_path, timestamp))

            cur_depth = len(cur_path.split(os.sep)) - my_path_depth
            if cur_depth > 5:  # Avoid going too deep
                break

    if ge_graph_files:
        sorted_ge_graph_files = [file for file, timestamp in sorted(ge_graph_files, key=lambda x: x[1])]
        return sorted_ge_graph_files
    return None


def get_unique_key(cur_dict, cur_key):
    split_sign, original_cur_key, cur_key_id = "#", cur_key, 0
    while cur_key in cur_dict:
        cur_key_id += 1
        cur_key = f"{original_cur_key}{split_sign}{cur_key_id}"
    return cur_key


def parse_pbtxt_to_dict(pbtxt_path):
    check_file_or_directory_path(pbtxt_path)
    with FileOpen(pbtxt_path, "r") as ff:
        contents = ff.read()

    result, cur_dict, superior_dicts, brackets_depth = [], {}, [], 0
    for cur_line in contents.split("\n"):
        cur_line = cur_line.strip()
        if len(cur_line) == 0:
            continue

        if " {" in cur_line:
            if brackets_depth == 0:
                cur_dict = {}
                superior_dicts = []
                result.append(cur_dict)
            cur_key = cur_line.split(" {")[0]
            cur_key = get_unique_key(cur_dict, cur_key)
            cur_dict[cur_key] = {}
            if len(superior_dicts) > brackets_depth:
                superior_dicts[brackets_depth] = cur_dict
            else:
                superior_dicts.append(cur_dict)
            cur_dict = cur_dict[cur_key]
            brackets_depth += 1
        elif ": " in cur_line:
            cur_key, cur_value = cur_line.split(": ")
            cur_key = get_unique_key(cur_dict, cur_key)
            cur_value = cur_value[1:-1] if cur_value.startswith('"') and cur_value.endswith('"') else cur_value
            cur_dict[cur_key] = cur_value
        elif "}" in cur_line:
            brackets_depth -= 1
            cur_dict = superior_dicts[brackets_depth]
    return result


def judge_single_or_multi_device(path):
    def is_time_directory(dir_name):
        if not os.path.isdir(os.path.join(path, dir_name)):
            return False
        return len(dir_name) == len(GE_DUMP_TIME_PATTERN) and str.isdigit(dir_name)

    # 获取指定目录下所有文件和文件夹
    entries = os.listdir(path)
    # 过滤出文件夹
    subdirs = [entry for entry in entries if os.path.isdir(os.path.join(path, entry))]
    if len(subdirs) > 1:
        return True

    time_dirs = [os.path.join(path, entry) for entry in entries if is_time_directory(entry)]
    if time_dirs:
        entries = os.listdir(time_dirs[0])
        subdirs = [entry for entry in entries if os.path.isdir(os.path.join(time_dirs[0], entry))]
        return len(subdirs) > 1

    return False


def _has_rank_directory(dump_dir):
    if not os.path.isdir(dump_dir):
        return False
    for entry in os.listdir(dump_dir):
        if os.path.isdir(os.path.join(dump_dir, entry)) and get_rank_id_from_torchair_data(entry) != -1:
            return True
    return False


def _validate_read_path(path):
    path_type = FileCheckConst.DIR if os.path.isdir(path) else FileCheckConst.FILE
    FileChecker(path, path_type, ability=FileCheckConst.READ_ABLE).common_check()


def gather_data_with_token_id_fx(data_path, token_dirs, rank_info_existed=False):
    for cur_path, dirs, _ in os.walk(data_path):
        if len(dirs) == 0:
            continue
        if all([len(ii) < MAX_TOKEN_LEN and str.isdigit(ii) for ii in dirs]):
            dirs = sorted(dirs, key=lambda xx: int(xx))
            token_dirs = [os.path.join(cur_path, dir_name) for dir_name in dirs]
            break

    if len(token_dirs) == 0:
        token_dirs.append(data_path)  # Just use data_path if found no token like dirs

    gathered_files_list = []

    if rank_info_existed:
        gathered_files = {}
        for token_dir in token_dirs:
            cur_token_id = os.path.basename(token_dir)
            cur_token_id = int(cur_token_id) + 1 if cur_token_id.isdigit() else 0
            file_names = [os.path.join(token_dir, f) for f in os.listdir(token_dir) if f.endswith(".npy")]
            gathered_files[cur_token_id] = file_names
        gathered_files_list.append(gathered_files)
        return gathered_files_list

    dump_dirs = {}
    for token_dir in token_dirs:
        cur_token_id = os.path.basename(token_dir)
        cur_token_id = int(cur_token_id) if cur_token_id.isdigit() else 0
        dump_dirs[cur_token_id] = sorted(
            [
                os.path.join(token_dir, d)
                for d in os.listdir(token_dir)
                if os.path.isdir(os.path.join(token_dir, d))
            ],
            key=lambda x: os.path.basename(x),
        )
    num_dumps = len(dump_dirs.get(1, None))
    for i in range(num_dumps):
        gathered_files = {}
        for cur_token_id, dumps in dump_dirs.items():
            dump_path = dumps[i]
            file_names = [os.path.join(dump_path, f) for f in os.listdir(dump_path) if f.endswith(".npy")]
            gathered_files[cur_token_id] = file_names
        gathered_files_list.append(gathered_files)
    return gathered_files_list


def gather_data_with_token_id(data_path, fx=False, rank_info_existed=False):
    token_dirs = []
    # Detect the deepest dir level where sub dirs are all digits, and regard as tokens level.
    if fx:
        return gather_data_with_token_id_fx(data_path, token_dirs, rank_info_existed)

    is_multi_device = False if rank_info_existed else judge_single_or_multi_device(data_path)

    for cur_path, dirs, _ in sorted(os.walk(data_path), key=lambda x: x[0]):
        if not dirs:
            token_dirs.append(cur_path)

    if len(token_dirs) == 0:
        token_dirs.append(data_path)  # Just use data_path if found no token like dirs

    gathered_files_list = []
    parent_dir_dict = {}
    for token_dir in token_dirs:
        if is_multi_device:
            parts = token_dir
            """
            token_dir格式如下：
            /home/dump/dump_20241114_113410/0/graph_1_0/1/4/
            dump路径+时间戳+device_id+子图名称+子图ID号+token_id
            对于多卡场景，应取device_id下相同的子图进行比较，
            此处parts=/home/dump/dump_20241114_113410/0
            """
            for _ in range(3):
                parts = os.path.dirname(parts)
            parent_dir = os.path.basename(parts)
        else:
            parent_dir = os.path.basename(os.path.dirname(token_dir))
        subdir = os.path.basename(token_dir)
        parent_id = int(parent_dir) if parent_dir.isdigit() else 0
        subdir_id = int(subdir) if subdir.isdigit() else 0
        if parent_id not in parent_dir_dict:
            parent_dir_dict[parent_id] = {}
        if subdir_id not in parent_dir_dict[parent_id]:
            parent_dir_dict[parent_id][subdir_id] = []
        for cur_path, _, file_names in os.walk(token_dir):
            file_names = [os.path.join(cur_path, file_name) for file_name in file_names]
            parent_dir_dict[parent_id][subdir_id].extend(file_names)
    parent_dir_dict = dict(sorted(parent_dir_dict.items()))
    for _, subdirs in parent_dir_dict.items():
        gathered_files_list.append(subdirs)
    return gathered_files_list


def init_ge_dump_data_from_bin_path(ge_dump_path):
    """
    For data like:
      1/Add.Add_2.44.6.1706596912161941,
      1/Cast.Cast_9.19.6.1706596911887829,
      1/ConcatV2D.ConcatV2.42.6.1706596912161117,

    Return dict:
      {1: {
            'Add_2': '1/Add.Add_2.44.6.1706596912161941',
            'Cast_9': '1/Cast.Cast_9.19.6.1706596911887829',
            'ConcatV2': '1/ConcatV2D.ConcatV2.42.6.1706596912161117',
      }}
    """
    gathered_files_list = gather_data_with_token_id(ge_dump_path)
    if not gathered_files_list:
        raise Exception("Cannot get ge dump data, because the gathered_files_list is empty.")

    dump_data_with_token_id_list = []
    for gathered_files in gathered_files_list:
        dump_data_with_token_id = {}
        for token_id, file_list in gathered_files.items():
            cur_dump_data = {}
            for file_name in sorted(file_list):
                if os.path.splitext(file_name)[-1] in DUMP_FILE_FILTER_SUFIX:
                    continue
                split_name = os.path.basename(file_name).split(".")
                if len(split_name) < 5:
                    logger.warning(f"invalid file name: {file_name}, should contain at least 4 '.'")
                    continue

                cur_op_name = ".".join(split_name[1:-3])
                if cur_op_name in cur_dump_data:
                    exists_file = cur_dump_data[cur_op_name]
                    exists_file_size = os.path.getsize(exists_file)
                    cur_file_size = os.path.getsize(file_name)
                    keep_one = file_name if cur_file_size > exists_file_size else exists_file
                    cur_dump_data[cur_op_name] = keep_one
                    logger.warning(
                        f"duplicated op name: {cur_op_name}."
                        f" [{os.path.basename(file_name)}, {os.path.basename(exists_file)}]."
                        f" Will keep the larger one {os.path.basename(keep_one)}."
                    )
                else:
                    cur_dump_data[cur_op_name] = file_name
            dump_data_with_token_id[token_id] = cur_dump_data
        dump_data_with_token_id_list.append(dump_data_with_token_id)
    return dump_data_with_token_id_list


def init_fx_dump_data_from_path(fx_dump_path, rank_info_existed=False):
    """
    For data like:
      1/mm-aten.mm.default.INPUT.0.20240125031118787351.npy,
      1/mm-aten.mm.default.INPUT.1.20240125031118787351.npy,
      1/mm-aten.mm.default.OUTPUT.0.20240125031118787351.npy,

    Return dict:
      {1: {'mm-aten.mm.default': {
        'input': [
          '1/mm-aten.mm.default.INPUT.0.20240125031118787351.npy',
          '1/mm-aten.mm.default.INPUT.1.20240125031118787351.npy',
        ],
        'output': ['1/mm-aten.mm.default.OUTPUT.0.20240125031118787351.npy']
      }}}
    """
    gathered_files_list = gather_data_with_token_id(fx_dump_path, fx=True, rank_info_existed=rank_info_existed)
    if not gathered_files_list:
        raise Exception("Cannot get fx dump data, because the gathered_files_list is empty.")

    dump_data_with_token_id_list = []
    for gathered_files in gathered_files_list:
        dump_data_with_token_id = {}
        for token_id, file_list in gathered_files.items():
            cur_dump_data = {}
            for file_path in sorted(file_list):
                if not file_path.endswith("npy"):
                    continue
                file_name = os.path.basename(file_path)
                is_input = ".INPUT." in file_name
                cur_op_name = file_name.split(".INPUT." if is_input else ".OUTPUT.")[0]
                cur_op_map = cur_dump_data.get(cur_op_name, {})
                cur_op_map.setdefault("input" if is_input else "output", []).append(file_path)
                cur_dump_data[cur_op_name] = cur_op_map
            if len(cur_dump_data) > 0:
                dump_data_with_token_id[token_id - 1] = cur_dump_data  # For FX data, token starts from 1, while GE is 0
        dump_data_with_token_id_list.append(dump_data_with_token_id)
    return dump_data_with_token_id_list


def compare_single_data(golden_path, my_path, token_id=0, golden_data=None, my_data=None):
    data_info = BasicDataInfo(golden_path, my_path, token_id)
    return fill_row_data(data_info, loaded_my_data=my_data, loaded_golden_data=golden_data)


# Comparing GE with FX
def filter_valid_fx_desc_tensor_info(desc_key, desc_value):
    """Valid one like: 'attr': {'key': '_fx_tensor_name', 'value': {'s': 'add_1-aten.add.Tensor.OUTPUT.0'}}"""
    if not (desc_key == "attr" or desc_key.startswith("attr#")) or not isinstance(desc_value, dict):
        return False
    if desc_value.get("key", None) != "_fx_tensor_name" or not isinstance(desc_value.get("value", None), dict):
        return False
    if not isinstance(desc_value.get("value", {}).get("s", None), str):
        return False
    return True


def get_all_ops_from_fusion_op(op_name, graph_map_dict, ge_dump_data):
    all_ops = []
    while len(op_name) > 0:
        cur_op_name = find_longest_name(op_name, graph_map_dict, ge_dump_data, ge_dump_data)
        if cur_op_name is None or cur_op_name not in graph_map_dict:
            logger.debug(f"Failed parsing ge op name: {cur_op_name}.Compare manually if required.")
            break
        all_ops.append(cur_op_name)
        op_name = op_name[len(cur_op_name):]
    return all_ops


def compare_ge_with_fx(graph_map, ge_dump_data, fx_dump_data, token_id=0):
    gathered_row_data = []
    graph_map_dict = {
        graph["op"]["name"]: graph["op"]
        for graph in graph_map
        if "op" in graph and "name" in graph["op"]
    }
    ge_dump_data = sort_ge_dump_data(ge_dump_data, graph_map)
    for op_name, my_path in ge_dump_data.items():
        all_ops = get_all_ops_from_fusion_op(op_name, graph_map_dict, ge_dump_data)
        if len(all_ops) == 1:
            op_info = graph_map_dict.get(all_ops[0])
            gathered_row_data.extend(compare_ge_with_fx_single_op(op_info, fx_dump_data, op_name, my_path, token_id))
        elif len(all_ops) > 1:
            first_op_info = graph_map_dict.get(all_ops[0])
            last_op_info = graph_map_dict.get(all_ops[-1])
            __args = [first_op_info, last_op_info, fx_dump_data, op_name, my_path, token_id]
            gathered_row_data.extend(compare_ge_with_fx_multiple_ops(*__args))
        else:
            op_type = os.path.basename(my_path).split(".")[0]
            if "Cast" in op_type or "TransData" in op_type:
                ge_inputs, ge_outputs = parse_torchair_dump_data(my_path)
                logger.debug(f"ge_inputs length: {len(ge_inputs)}")
                logger.debug(f"ge_outputs length:, {len(ge_outputs)}")
                gathered_row_data.extend(compare_specials_private_ops(ge_inputs, ge_outputs, token_id, my_path))

    return gathered_row_data


def compare_ge_with_fx_single_op(op_info, fx_dump_data, op_name, my_path, token_id=0):
    gathered_row_data = []
    for op_key, op_value in op_info.items():
        if not (op_key == "output_desc" or op_key.startswith("output_desc#")) or not isinstance(op_value, dict):
            continue
        for out_key, out_value in op_value.items():
            if not filter_valid_fx_desc_tensor_info(out_key, out_value):
                continue
            fx_tensor_name = out_value.get("value", {}).get("s", None)
            if fx_tensor_name.split(".")[-2] == "OUTPUT":
                fx_tensor_name = ".".join(fx_tensor_name.split(".")[:-2])
            if fx_tensor_name not in fx_dump_data:
                logger.warning(f"FX data missing, GE tensor name: {op_name}, FX tensor name: {fx_tensor_name}")
                continue

            ge_inputs, ge_outputs = parse_torchair_dump_data(my_path)
            fx_inputs = fx_dump_data.get(fx_tensor_name, {}).get("input", [])
            fx_outputs = fx_dump_data.get(fx_tensor_name, {}).get("output", [])
            logger.debug(f"ge_inputs length: {len(ge_inputs)}, fx_inputs length:, {len(fx_inputs)}")
            logger.debug(f"ge_outputs length: {len(ge_outputs)}, fx_outputs length:, {len(fx_outputs)}")
            gathered_row_data = compare_ops((fx_inputs, fx_outputs), (ge_inputs, ge_outputs), token_id, my_path)

    return gathered_row_data


def compare_ge_with_fx_multiple_ops(first_op_info, last_op_info, fx_dump_data, op_name, my_path, token_id):
    gathered_row_data = []
    gathered_row_data.extend(
        compare_ge_with_fx_multiple_ops_details(first_op_info, fx_dump_data, op_name, my_path, "input", token_id)
    )
    gathered_row_data.extend(
        compare_ge_with_fx_multiple_ops_details(last_op_info, fx_dump_data, op_name, my_path, "output", token_id)
    )
    return gathered_row_data


def compare_ge_with_fx_multiple_ops_details(op_info: dict, *args):
    fx_dump_data, op_name, my_path, input_or_output, token_id = args
    gathered_row_data = []
    for op_key, op_value in op_info.items():
        if not (op_key == "output_desc" or op_key.startswith("output_desc#")) or not isinstance(op_value, dict):
            continue
        for out_key, out_value in op_value.items():
            if not filter_valid_fx_desc_tensor_info(out_key, out_value):
                continue
            fx_tensor_name = out_value.get("value", {}).get("s", None)
            if fx_tensor_name.split(".")[-2] == "OUTPUT":
                fx_tensor_name = ".".join(fx_tensor_name.split(".")[:-2])
            if fx_tensor_name not in fx_dump_data:
                logger.warning(f"FX data missing, GE tensor name: {op_name}, FX tensor name: {fx_tensor_name}")
                continue
            ge_inputs, ge_outputs = parse_torchair_dump_data(my_path)
            fx_inputs_or_outputs = fx_dump_data.get(fx_tensor_name, {}).get(input_or_output, [])
            ge_input_or_output_path = ""
            ge_inputs_or_outputs = []
            if input_or_output == "input":
                logger.debug(f"ge_inputs length: {len(ge_inputs)}, fx_inputs length:, {len(fx_inputs_or_outputs)}")
                ge_input_or_output_path = "inputs"
                ge_inputs_or_outputs = ge_inputs
            elif input_or_output == "output":
                logger.debug(f"ge_outputs length: {len(ge_outputs)}, fx_outputs length:, {len(fx_inputs_or_outputs)}")
                ge_input_or_output_path = "outputs"
                ge_inputs_or_outputs = ge_outputs
            for cur_id, (fx_input_or_output, ge_input_or_output) in enumerate(
                zip(fx_inputs_or_outputs, ge_inputs_or_outputs)
            ):
                cur_ge_data = "{},{},{}".format(my_path, ge_input_or_output_path, cur_id)
                row_data = compare_single_data(fx_input_or_output, cur_ge_data, token_id, my_data=ge_input_or_output)
                gathered_row_data.append(row_data)

    return gathered_row_data


def compare_specials_private_ops(ge_inputs, ge_outputs, token_id, my_path):
    gathered_row_data = []
    for cur_id, (ge_input, ge_output) in enumerate(zip(ge_inputs, ge_outputs)):
        cur_ge_input_data = f"{my_path},inputs,{cur_id}"
        cur_ge_output_data = f"{my_path},outputs,{cur_id}"
        row_data = compare_single_data(cur_ge_input_data, cur_ge_output_data, token_id, ge_input, ge_output)
        gathered_row_data.append(row_data)

    return gathered_row_data


def compare_ops(fx_tuple, ge_tuple, token_id, my_path):
    gathered_row_data = []
    for cur_id, (fx_input, ge_input) in enumerate(zip(fx_tuple[0], ge_tuple[0])):
        cur_ge_data = f"{my_path},inputs,{cur_id}"
        row_data = compare_single_data(fx_input, cur_ge_data, token_id, my_data=ge_input)
        gathered_row_data.append(row_data)
    for cur_id, (fx_output, ge_output) in enumerate(zip(fx_tuple[1], ge_tuple[1])):
        cur_ge_data = f"{my_path},outputs,{cur_id}"
        row_data = compare_single_data(fx_output, cur_ge_data, token_id, my_data=ge_output)
        gathered_row_data.append(row_data)

    return gathered_row_data


# Comparing fused GE with GE
def get_all_op_input_names(op_info):
    inputs = [vv for kk, vv in op_info.items() if kk == "input" or kk.startswith("input#")]
    return [":".join(ii.split(":")[:-1]) for ii in inputs]


def find_longest_name(op_name, op_map, fused_ge_dump_data, ge_dump_data):
    if op_name in op_map:
        return op_name
    op_name_len = len(op_name)
    for idx in range(1, op_name_len):
        cur_op_name = op_name[:-idx]
        if cur_op_name in op_map:
            return cur_op_name
        if cur_op_name in fused_ge_dump_data or cur_op_name in ge_dump_data:
            return None  # op_name in dump data but not op_map, abandon
    return None


def gather_fused_op_data(fused_op_name, op_map, fused_ge_dump_data, ge_dump_data):
    gathered_input_names, gathered_inputs, gatherd_input_pathes, gathered_ops = [], [], [], []
    output_path, op_outputs = None, []
    while len(fused_op_name) > 0:
        cur_op_name = find_longest_name(fused_op_name, op_map, fused_ge_dump_data, ge_dump_data)
        if cur_op_name is None or cur_op_name not in op_map:
            logger.warning(f"Failed parsing fused op name: {fused_op_name}. Compare manually if required.")
            break
        cur_input_names = get_all_op_input_names(op_map[cur_op_name])

        if cur_op_name in ge_dump_data:
            cur_path = ge_dump_data[cur_op_name]
            op_inputs, op_outputs = parse_torchair_dump_data(cur_path)
            min_inputs_len = min(len(cur_input_names), len(op_inputs))
            cur_input_names, op_inputs = cur_input_names[:min_inputs_len], op_inputs[:min_inputs_len]
            input_pathes = [",".join([cur_path, "inputs", str(idx)]) for idx in range(min_inputs_len)]
            output_path = cur_path  # Till get the last op path
        else:
            logger.warning(
                f"No dump data for op: {cur_op_name}. Seldom should this happen. Input data matching may be incorrect."
            )
            empty_data = np.array([], dtype="float32")
            op_inputs = [empty_data] * len(cur_input_names)
            input_pathes = [""] * len(cur_input_names)

        gathered_input_names.extend(cur_input_names)
        gathered_ops.append(cur_op_name)
        gathered_inputs.extend(op_inputs)
        gatherd_input_pathes.extend(input_pathes)
        fused_op_name = fused_op_name[len(cur_op_name):]

    filtered_input_names, filtered_inputs, filtered_input_pathes = [], [], []
    for input_name, inputs, input_path in zip(gathered_input_names, gathered_inputs, gatherd_input_pathes):
        if input_name not in gathered_ops:
            filtered_input_names.append(input_name)
            filtered_input_pathes.append(input_path)
            filtered_inputs.append(inputs)
    return (filtered_inputs, filtered_input_pathes), (op_outputs, output_path)  # op_outputs is just the last op output


def compare_ge_with_ge(graph_map, fused_ge_dump_data, ge_dump_data, token_id=0):
    graph_map_dict = {ii["op"]["name"]: ii["op"] for ii in graph_map if "op" in ii and "name" in ii["op"]}
    fused_ge_dump_data = sort_ge_dump_data(fused_ge_dump_data, graph_map)
    gathered_row_data = []
    for op_name, my_path in fused_ge_dump_data.items():
        is_fused_op = os.path.basename(my_path).startswith(FUSION_OP_TYPE)
        if is_fused_op:
            (golden_inputs, golden_input_pathes), (golden_outputs, golden_output_path) = gather_fused_op_data(
                op_name, graph_map_dict, fused_ge_dump_data, ge_dump_data
            )
        elif op_name in ge_dump_data:
            golden_path = ge_dump_data[op_name]
            golden_inputs, golden_outputs = parse_torchair_dump_data(golden_path)
            golden_input_pathes = [golden_path] * len(golden_inputs)
            golden_output_path = golden_path
        else:
            logger.warning(f"Golden data missing, My tensor name: {op_name}")
            continue

        my_inputs, my_outputs = parse_torchair_dump_data(my_path)
        logger.debug(f"golden_inputs length: {len(golden_inputs)}, my_inputs length:, {len(my_inputs)}")
        logger.debug(f"golden_outputs length: {len(golden_outputs)}, my_outputs length:, {len(my_outputs)}")

        for cur_id, (golden_input, my_input, golden_input_path) in enumerate(
            zip(golden_inputs, my_inputs, golden_input_pathes)
        ):
            cur_ge_data = f"{my_path},inputs,{cur_id}"
            if ",inputs," not in golden_output_path:
                golden_output_path = f"{golden_output_path},inputs,{cur_id}"
            row_data = compare_single_data(
                golden_input_path, cur_ge_data, token_id, golden_data=golden_input, my_data=my_input
            )
            gathered_row_data.append(row_data)
        for cur_id, (golden_output, my_output) in enumerate(zip(golden_outputs, my_outputs)):
            cur_ge_data = f"{my_path},outputs,{cur_id}"
            golden_output_path = f"{golden_output_path},outputs,{cur_id}"
            row_data = compare_single_data(
                golden_output_path, cur_ge_data, token_id, golden_data=golden_output, my_data=my_output
            )
            gathered_row_data.append(row_data)
    return gathered_row_data


def sort_ge_dump_data(dump_data, graph_map):
    graph_map_sort = {graph["op"]["name"]: id for id, graph in enumerate(graph_map)}
    sort_ops_list = sorted(dump_data, key=lambda x: graph_map_sort.get(x, -1))
    ge_dump_data = OrderedDict((op_name, dump_data[op_name]) for op_name in sort_ops_list)
    return ge_dump_data


def sort_by_timestamp(gathered_row_data):
    """
    gathered_row_data为保存比对结果的列表，列表里每个元素都是一个字典
    1. 取每个字典的'my_data_path'字段x['my_data_path']，该字段的值如：OpType.OpName.12.7.1734070081497686,inputs,0
    2. 按'.'分隔字符串x['my_data_path'].split('.') —> ['OpType', 'OpName', '12', '7', '1734070081497686,inputs,0']
    3. x['my_data_path'].split('.')[-1]取最后一个元素为'1734070081497686,inputs,0'，该值包含时间戳（1734070081497686）和输入输出信息（inputs,0）
    4. 按'1734070081497686,inputs,0'对比对结果重新排序得到sorted_gathered_row_data列表
    5. 返回排序后的结果
    """
    sorted_gathered_row_data = sorted(gathered_row_data, key=lambda x: x['my_data_path'].split('.')[-1])
    return sorted_gathered_row_data


# Main entrance
def acc_compare(golden_path, my_path, output_path='./', rank_id=None, rank_info_existed=False):
    set_msaccucmp_path_from_cann()

    if not get_torchair_ge_graph_path(my_path):
        raise Exception("Can not get ge graph, Please check whether the input path contains graph.")

    if rank_info_existed:
        if rank_id is None:
            golden_data_ranks = set()
            my_data_ranks = set()

            for subdir in os.listdir(golden_path):
                rank_id = get_rank_id_from_torchair_data(subdir)
                if os.path.isdir(os.path.join(golden_path, subdir)) and rank_id != -1:
                    golden_data_ranks.add(rank_id)

            for subdir in os.listdir(my_path):
                rank_id = get_rank_id_from_torchair_data(subdir)
                if os.path.isdir(os.path.join(my_path, subdir)) and rank_id != -1:
                    my_data_ranks.add(rank_id)

            compared_ranks = list(golden_data_ranks & my_data_ranks)
            if not compared_ranks:
                raise Exception("No common rank data in golden_path and my_path.")
        else:
            compared_ranks = [rank_id]
    else:
        compared_ranks = [-1]

    args = [(golden_path, my_path, output_path, rid) for rid in compared_ranks]
    # If only a single rank needs comparison, run synchronously so that any exceptions
    # raised inside `acc_compare_once` are propagated directly.
    if len(args) == 1:
        acc_compare_once(args[0])
    else:
        # Use multiprocessing for multiple ranks but make sure to propagate exceptions
        processes_pool = Pool(min(len(args), int(cpu_count() * 1.3)))
        async_results = [processes_pool.apply_async(save_compare_once, (arg,)) for arg in args]
        processes_pool.close()
        # Ensure that exceptions in worker processes are not silenced. Calling `get()` will
        # re-raise any exception occurred in the worker process in the main process.
        for res in async_results:
            res.get()
        processes_pool.join()

    return output_path


def save_compare_once(args):
    try:
        return acc_compare_once(args)
    except Exception as e:
        error_msg = f"Error in acc_compare_once: {str(e)}"
        raise ValueError(error_msg) from e


def compare_torchair_mode(args):
    """
    Entry point used by the CLI to trigger torchair accuracy compare.
    """
    my_path = os.path.realpath(args.target_path)
    golden_path = os.path.realpath(args.golden_path)
    _validate_read_path(my_path)
    _validate_read_path(golden_path)

    rank_arg = getattr(args, "rank", None)
    rank_id = None
    if rank_arg is not None:
        rank_str = str(rank_arg).strip()
        if not rank_str.isdigit():
            logger.error("Argument --rank only supports a single integer when mode=='torchair'.")
            raise CompareException(CompareException.INVALID_PARAM_ERROR, "Invalid rank parameter for torchair mode.")
        rank_id = int(rank_str)

    rank_info_existed = _has_rank_directory(my_path)
    if not rank_info_existed and rank_id is not None:
        logger.warning('The directory structure of torchair data is old, the rank parameter will not take effect.')
    logger.info(f"[compare_torchair] start comparing, golden_path: {golden_path}, target_path: {my_path}")
    return acc_compare(golden_path, my_path, args.output_path, rank_id, rank_info_existed)


def acc_compare_once(*args):
    dir_of_golden_path, dir_of_my_path, output_path, rank_id = args[0]
    if rank_id != -1:
        subdirs = []
        for subdir in os.listdir(dir_of_golden_path):
            is_dir = os.path.isdir(os.path.join(dir_of_golden_path, subdir))
            if is_dir and subdir.startswith('worldsize') and subdir.endswith(f'rank{rank_id}'):
                subdirs.append(subdir)
        if not subdirs:
            raise Exception(f'Can not get golden data in rank {rank_id}')
        golden_path = os.path.join(dir_of_golden_path, subdirs[-1])

        subdirs = []
        for subdir in os.listdir(dir_of_my_path):
            is_dir = os.path.isdir(os.path.join(dir_of_my_path, subdir))
            if is_dir and subdir.startswith('worldsize') and subdir.endswith(f'rank{rank_id}'):
                subdirs.append(subdir)
        if not subdirs:
            raise Exception(f'Can not get my data in rank {rank_id}')
        my_path = os.path.join(dir_of_my_path, subdirs[-1])
    else:
        golden_path = dir_of_golden_path
        my_path = dir_of_my_path

    ge_graph_path = get_torchair_ge_graph_path(dir_of_my_path, rank_id)

    if not ge_graph_path:
        raise Exception("Can not get ge graph, Please check whether the input path contains graph.")

    logger.info(f"[compare_torchair], golden_path: {golden_path}, my_path: {my_path}, ge_graph_path: {ge_graph_path}")

    graph_map_list = []
    for path in ge_graph_path:
        graph_map_list.append(parse_pbtxt_to_dict(path))

    my_dump_data_list = init_ge_dump_data_from_bin_path(my_path)

    is_golden_fx = get_torchair_ge_graph_path(dir_of_golden_path) is None
    if is_golden_fx:
        logger.info("Comparing GE with FX")
        golden_dump_data_list = init_fx_dump_data_from_path(golden_path, rank_id != -1)
    else:
        logger.info("Comparing GE with GE")
        golden_dump_data_list = init_ge_dump_data_from_bin_path(golden_path)

    graph_map_list_len = len(graph_map_list)
    for i in range(graph_map_list_len):
        logger.info(f"All token ids in my_dump_data: {my_dump_data_list[i].keys()}")
        logger.info(f"All token ids in golden_dump_data: {golden_dump_data_list[i].keys()}")
        graph_map = graph_map_list[i]
        my_dump_data = my_dump_data_list[i]
        golden_dump_data = golden_dump_data_list[i]

        gathered_row_data = []
        for token_id in my_dump_data:
            if token_id not in golden_dump_data:
                logger.warning(f"My token_id {token_id} not found in golden dump data")
                continue
            logger.info(f"Comparing token_id: {token_id}")
            if is_golden_fx:
                row_data = compare_ge_with_fx(graph_map, my_dump_data[token_id], golden_dump_data[token_id], token_id)
            else:
                row_data = compare_ge_with_ge(graph_map, my_dump_data[token_id], golden_dump_data[token_id], token_id)
            gathered_row_data.extend(row_data)
        sorted_gathered_row_data = sort_by_timestamp(gathered_row_data)
        save_compare_result_to_csv(sorted_gathered_row_data, output_path, rank_id=rank_id)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
import math

from msprobe.core.common.const import Const


def postprocess_pass(data_items, name2item):
    backward_pass(data_items, name2item)
    renumber_index_pass(data_items, "ParallelTransformer", "layers")


def backward_pass(data_items, name2item):
    # 处理反向数据，反向无栈信息，沿用正向数据栈信息
    for data_item in data_items:
        data_name_list = data_item.data_name.split(Const.SEP)
        if not data_name_list:
            continue
        if Const.BACKWARD in data_name_list[Const.SCOPE_DIRECTION_INDEX:]:
            data_name_list[Const.SCOPE_DIRECTION_INDEX:] = [
                s.replace(Const.BACKWARD, Const.FORWARD)
                for s in data_name_list[Const.SCOPE_DIRECTION_INDEX:]
            ]
            forward_name = Const.SEP.join(data_name_list)
            forward_item = name2item.get(forward_name, None)
            if not forward_item:
                continue
            data_item.stack_scope = forward_item.stack_scope
            data_item.full_scope = forward_item.full_scope
            data_item.layer_scope = forward_item.layer_scope


def extract_next_item_last_number(data, prefix, default_result=None):
    result = default_result
    match = re.search(rf"^{re.escape(prefix)}\.(\S+?)(?:\.|$)", data)
    if match:
        next_item = match.group(1)
        numbers = re.findall(r"\d+", next_item)
        if numbers:
            result = int(numbers[-1])
    return result


def replace_next_item_index(full_scope, prefix, index):
    if math.isinf(index):
        return full_scope
    prefix_pattern = rf"^{re.escape(prefix)}\."
    result = full_scope
    match = re.search(rf"{prefix_pattern}(\S+?)(?:\.|$)", full_scope)
    if match:
        next_item = match.group(1)
        pattern = rf"{prefix_pattern}{re.escape(next_item)}"
        result = re.sub(pattern, f"{prefix}.{index}", full_scope, count=1)
    return result


def renumber_index_pass(data_items, type_name, suffix=None):
    """
    该函数为解决并行切分场景中编号不一致的比对问题。例如在MindSpore中ParallelTransformer层的PP切分场景，
    MindSpore中的layers的成员编号是全局的，而在PyTorch中编号为局部的。
    为适配此种场景，对指定层的索引进行重新编号，以确保在后续处理阶段序号对齐。
    """
    prefix_dict = {}  # 保存类型为type_name的前缀和最小编号的映射
    for data_item in data_items:
        if data_item.type_name == type_name:
            prefix = f"{data_item.full_scope}.{suffix}" if suffix else data_item.layer_scope
            prefix_dict[prefix] = math.inf

    # 计算前缀对应的最小编号
    for prefix in prefix_dict:
        for data_item in data_items:
            res = extract_next_item_last_number(data_item.full_scope, prefix, math.inf)
            prefix_dict[prefix] = min(prefix_dict[prefix], res)

    # 重新编号
    for prefix, min_index in prefix_dict.items():
        for data_item in data_items:
            full_scope = data_item.full_scope
            abs_index = extract_next_item_last_number(data_item.full_scope, prefix, math.inf)
            rel_index = abs_index - min_index
            full_scope = replace_next_item_index(full_scope, prefix, rel_index)
            data_item.full_scope = full_scope



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class is used to om fusion parser.
"""
import itertools
import json
from typing import Union, List

import numpy as np

from msprobe.core.common.log import logger
from msprobe.infer.offline.compare.msquickcmp.common import utils
from msprobe.infer.offline.compare.msquickcmp.common.dynamic_argument_bean import DynamicArgumentEnum
from msprobe.infer.offline.compare.msquickcmp.common.utils import AccuracyCompareException
from msprobe.infer.utils.util import load_file_to_read_common_check
from msprobe.infer.utils.file_open_check import ms_open
from msprobe.infer.utils.constants import TENSOR_MAX_SIZE


GRAPH_OBJECT = "graph"
OP_OBJECT = "op"
NAME_OBJECT = "name"
TYPE_OBJECT = "type"
INPUT_DESC_OBJECT = "input_desc"
INPUT_OBJECT = "input"
ATTR_OBJECT = "attr"
SHAPE_OBJECT = "shape"
SHAPE_RANGE_OBJECT = "shape_range"
DIM_OBJECT = "dim"
DATA_OBJECT = "Data"
NET_OUTPUT_OBJECT = "NetOutput"
ATC_CMDLINE_OBJECT = "atc_cmdline"
INPUT_SHAPE = "--input_shape"
INPUT_SHAPE_RANGE = "--input_shape_range"
LIST_LIST_INT_OBJECT = 'list_list_int'
LIST_LIST_I_OBJECT = 'list_list_i'
LIST_I_OBJECT = 'list_i'
LIST_OBJECT = 'list'
KEY_OBJECT = "key"
VALUE_OBJECT = "value"
SUBGRAPH_NAME = 'subgraph_name'
S_OBJECT = "s"
DTYPE_OBJECT = "dtype"
DTYPE_MAP = {
    "DT_FLOAT": np.float32,
    "DT_FLOAT16": np.float16,
    "DT_DOUBLE": np.float64,
    "DT_INT8": np.int8,
    "DT_INT16": np.int16,
    "DT_INT32": np.int32,
    "DT_INT64": np.int64,
    "DT_UINT8": np.uint8,
    "DT_UINT16": np.uint16,
    "DT_UINT32": np.uint32,
    "DT_UINT64": np.uint64,
    "DT_BOOL": np.bool_,
}
OUT_NODES_NAME = "attr_model_out_nodes_name"
AIPP_CONFIG_PATH = "aipp_config_path"
LAYOUT_OBJECT = "layout"
# special ops
SPECIAL_OPS_TYPE = ("Cast", "TransData")


class OmParser(object):
    """
    This class is used to parse om model.
    """

    def __init__(self, output_json_path):
        self.json_object = self._load_json_file(output_json_path)
        self.subgraph_name = self.get_sub_graph_name()
        self.shape_range = self._is_input_shape_range()
        self.contain_negative_1 = False
        self.special_op_attr = self._parse_special_op_attr()

    @staticmethod
    def _load_json_file(json_file_path):
        """
        Function Description:
            load json file
        Parameter:
            json_file_path: json file path
        Return Value:
            json object
        Exception Description:
            when invalid json file path throw exception
        """
        try:
            json_file_path = load_file_to_read_common_check(json_file_path)
            with ms_open(json_file_path, "r", max_size=TENSOR_MAX_SIZE) as input_file:
                try:
                    return json.load(input_file)
                except Exception as exc:
                    logger.error(f"Load Json {json_file_path} failed, {str(exc)}")
                    raise AccuracyCompareException(utils.ACCURACY_COMPARISON_PARSER_JSON_FILE_ERROR) from exc
        except IOError as input_file_open_except:
            logger.error('Failed to open"' + json_file_path + '", ' + str(input_file_open_except))
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_OPEN_FILE_ERROR) from input_file_open_except

    @staticmethod
    def _get_prefix(input_obj):
        return input_obj.split(':')[0]

    @staticmethod
    def _parse_net_output_node_attr(operator):
        net_output_info = {}
        if INPUT_DESC_OBJECT in operator:
            input_index = 0
            for input_object in operator.get(INPUT_DESC_OBJECT):
                shape = []
                data_type = DTYPE_MAP.get(input_object.get(DTYPE_OBJECT))
                if not input_object.get(SHAPE_OBJECT):
                    # no shape info, assumed to be scalar
                    net_output_info[input_index] = [data_type, [1]]
                    continue
                for num in input_object.get(SHAPE_OBJECT).get(DIM_OBJECT):
                    shape.append(num)
                net_output_info[input_index] = [data_type, shape]
                input_index += 1
        return net_output_info

    def get_shape_size(self):
        """
        Get shape size for input
        """
        input_desc_array = self._get_data_input_desc()
        # extracts the input shape value
        return self._process_inputs(input_desc_array)

    def get_shape_list(self):
        """
        Get shape list for input
        """
        input_desc_array = self._get_data_input_desc()
        # extracts the input shape value
        return self._process_inputs_to_list(input_desc_array)

    def get_net_output_count(self):
        """
        Get net output count
        """
        count = 0
        is_dynamic_scenario, dym_arg = self.get_dynamic_scenario_info()
        if not is_dynamic_scenario or dym_arg is DynamicArgumentEnum.DYM_SHAPE:
            op_iterator = self._gen_operator_list()
        else:
            op_iterator = self._gen_operator_list_from_subgraph()
        for operator in op_iterator:
            if NET_OUTPUT_OBJECT == operator.get(TYPE_OBJECT) and INPUT_DESC_OBJECT in operator:
                count += len(operator.get(INPUT_DESC_OBJECT))
        return count

    def get_atc_cmdline(self):
        for attr in self.json_object.get(ATTR_OBJECT):
            if KEY_OBJECT in attr and attr.get(KEY_OBJECT) == ATC_CMDLINE_OBJECT:
                if VALUE_OBJECT in attr and S_OBJECT in attr.get(VALUE_OBJECT):
                    atc_cmd = attr.get(VALUE_OBJECT).get(S_OBJECT)
                    return self._replace_input_shape_with_input_shape_range_in_atc_cmd_if_dynamic(atc_cmd)
        return ''

    def get_expect_net_output_name(self):
        """
        Get the expected output tensor corresponding to Net_output.
        """
        net_output_names = []
        if ATTR_OBJECT not in self.json_object:
            return {}
        for attr in self.json_object.get(ATTR_OBJECT):
            if not (KEY_OBJECT in attr and attr.get(KEY_OBJECT) == OUT_NODES_NAME):
                continue
            if not (VALUE_OBJECT in attr and LIST_OBJECT in attr.get(VALUE_OBJECT)):
                continue
            list_object = attr.get(VALUE_OBJECT).get(LIST_OBJECT)
            if S_OBJECT in list_object:
                net_output_names = list_object.get(S_OBJECT)
        return dict(enumerate(net_output_names))

    def get_net_output_data_info(self, dump_data_path):
        """
        get_net_output_data_info
        """
        net_output_list = []
        for operator in self._gen_operator_list():
            if NET_OUTPUT_OBJECT == operator.get(TYPE_OBJECT):
                net_output_list.append(operator)
        if len(net_output_list) == 1:
            return self._parse_net_output_node_attr(net_output_list[0])
        # if it's dynamic batch scenario, the net output node should be identified by batch index
        _, scenario = self.get_dynamic_scenario_info()
        if scenario in [DynamicArgumentEnum.DYM_BATCH, DynamicArgumentEnum.DYM_DIMS]:
            if not dump_data_path:
                for operator in net_output_list:
                    return self._parse_net_output_node_attr(operator)
            cur_batch_index = utils.get_batch_index(dump_data_path)
            for operator in net_output_list:
                batch_index_in_operator = utils.get_batch_index_from_name(operator.get(NAME_OBJECT))
                if cur_batch_index == batch_index_in_operator:
                    return self._parse_net_output_node_attr(operator)
        logger.error("get npu output node info failed.")
        raise AccuracyCompareException(utils.ACCURACY_COMPARISON_PARSER_JSON_FILE_ERROR)

    def get_dynamic_scenario_info(self):
        atc_cmd = self.get_atc_cmdline()
        for dym_arg in DynamicArgumentEnum:
            if dym_arg.value.atc_arg in atc_cmd:
                return True, dym_arg
        return False, None

    def get_aipp_config_content(self) -> Union[str, List]:
        aipp_configs = []
        for graph in self.json_object.get(GRAPH_OBJECT):
            for op in graph.get(OP_OBJECT):
                for attr in op.get(ATTR_OBJECT):
                    if KEY_OBJECT in attr and attr.get(KEY_OBJECT) == AIPP_CONFIG_PATH:
                        aipp_configs.append(attr.get(VALUE_OBJECT).get(S_OBJECT))
        if aipp_configs:
            return aipp_configs
        else:
            return []

    def get_sub_graph_name(self):
        subgraph_name = []
        for graph in self.json_object.get(GRAPH_OBJECT):
            for operator in graph.get(OP_OBJECT):
                if SUBGRAPH_NAME in operator:
                    subgraph_name += operator.get(SUBGRAPH_NAME)
        return subgraph_name

    def _replace_input_shape_with_input_shape_range_in_atc_cmd_if_dynamic(self, atc_cmd):
        # `--input_shape_range` is deprecated, and replaced with `--input_shape` in atc.
        # but `--input_shape_range` is currently used in parsing model as dynamic shape one,
        # It's more convinient for using and doesn't matter in comparing process.
        # Thus `--input_shape` args is converted to `--input_shape_range` in
        # this function for further comparing usage.
        atc_cmd_split = []
        for cmd_token in atc_cmd.split():
            atc_cmd_split += cmd_token.split("=")  # Could be in format like "--input_shape=input:1,3,224,224"

        # Return if any dynamic args already involved
        for dym_arg in DynamicArgumentEnum:
            if dym_arg.value.atc_arg in atc_cmd_split:
                return atc_cmd

        # Check if `--input_shape` in atc_cmd line. Return if not
        input_shape_pos = -1
        for pos, cmd_token in enumerate(atc_cmd_split):
            if cmd_token == INPUT_SHAPE:
                input_shape_pos = pos

        if input_shape_pos == -1 or len(atc_cmd_split) <= input_shape_pos + 1:
            return atc_cmd

        # Check if dynamic `--input_shape`
        input_shape_args = atc_cmd_split[input_shape_pos + 1]
        if "-1" not in input_shape_args and "~" not in input_shape_args:
            return atc_cmd

        # replace `--input_shape` with `--input_shape_range` if dynamic
        input_shape_range_args_split = []
        for inputs in input_shape_args.split(';'):
            inputs_split = inputs.split(':')
            input_name, input_shape = ":".join(inputs_split[:-1]), inputs_split[-1]
            input_shape_range_args_split.append('{}:[{}]'.format(input_name, input_shape))
        input_shape_range_args = ';'.join(input_shape_range_args_split)

        atc_cmd_with_range_split = atc_cmd_split[:input_shape_pos]
        atc_cmd_with_range_split += [INPUT_SHAPE_RANGE, input_shape_range_args]
        atc_cmd_with_range_split += atc_cmd_split[input_shape_pos + 2:]
        atc_cmd_with_range = " ".join(atc_cmd_with_range_split)

        if not self.shape_range:  # Only print once, just if shape_range is False
            logger.info(
                f"Convert atc arg '{INPUT_SHAPE} {input_shape_args}' -> '{INPUT_SHAPE_RANGE} {input_shape_range_args}'"
            )
        self.shape_range = True
        return atc_cmd_with_range

    def _gen_operator_list(self):
        _, scenario = self.get_dynamic_scenario_info()
        for graph in self.json_object.get(GRAPH_OBJECT):
            if graph.get(NAME_OBJECT) in self.subgraph_name and scenario not in [
                DynamicArgumentEnum.DYM_BATCH,
                DynamicArgumentEnum.DYM_DIMS,
            ]:
                continue
            for operator in graph.get(OP_OBJECT):
                yield operator

    def _gen_operator_list_from_subgraph(self):
        for graph in self.json_object.get(GRAPH_OBJECT):
            if graph.get(NAME_OBJECT) in self.subgraph_name:
                for operator in graph.get(OP_OBJECT):
                    yield operator
                return

    def _get_data_input_desc(self):
        input_desc_list = []
        for operator in self._gen_operator_list():
            if DATA_OBJECT == operator.get(TYPE_OBJECT):
                if len(operator.get(INPUT_DESC_OBJECT)) != 0:
                    for item in operator.get(INPUT_DESC_OBJECT):
                        input_desc_list.append(item)
        return input_desc_list

    def _parse_special_op_attr(self):
        special_op_attr = {}
        for operator in self._gen_operator_list():
            if operator.get(TYPE_OBJECT) in SPECIAL_OPS_TYPE:
                special_op_attr[operator.get(NAME_OBJECT)] = operator.get(INPUT_OBJECT)
        return special_op_attr

    def _is_input_shape_range(self):
        if ATTR_OBJECT not in self.json_object:
            return False
        for attr in self.json_object.get(ATTR_OBJECT):
            if KEY_OBJECT in attr and attr.get(KEY_OBJECT) == ATC_CMDLINE_OBJECT:
                if VALUE_OBJECT in attr and S_OBJECT in attr.get(VALUE_OBJECT):
                    if INPUT_SHAPE_RANGE in attr.get(VALUE_OBJECT).get(S_OBJECT):
                        return True
        return False

    def _get_shape_list(self, list_list_int_object, shape_list):
        if LIST_LIST_I_OBJECT in list_list_int_object:
            for list_list_i in list_list_int_object.get(LIST_LIST_I_OBJECT):
                if LIST_I_OBJECT in list_list_i:
                    list_i = list_list_i.get(LIST_I_OBJECT)
                    if -1 in list_i:
                        self.contain_negative_1 = True
                        return
                    if len(list_i) != 2:
                        continue
                    shape_list.append(list(range(list_i[0], list_i[1] + 1)))
        return

    def _get_range_shape_size_list(self, input_object):
        range_shape_size_list = []
        if ATTR_OBJECT not in input_object:
            return range_shape_size_list
        shape_list = []
        for attr in input_object.get(ATTR_OBJECT):
            if KEY_OBJECT in attr and attr.get(KEY_OBJECT) == SHAPE_RANGE_OBJECT:
                if VALUE_OBJECT in attr and attr.get(VALUE_OBJECT) and LIST_LIST_INT_OBJECT in attr.get(VALUE_OBJECT):
                    list_list_int_object = attr.get(VALUE_OBJECT).get(LIST_LIST_INT_OBJECT)
                    self._get_shape_list(list_list_int_object, shape_list)
                    if self.contain_negative_1:
                        return []
        shape_list_all = list(itertools.product(*shape_list))
        for item in shape_list_all:
            item_sum = 1
            for num in item:
                item_sum *= num
            range_shape_size_list.append(item_sum)
        return range_shape_size_list

    def _process_inputs(self, input_desc_array):
        value = []
        for input_object in input_desc_array:
            if SHAPE_OBJECT not in input_object:
                value.append(0)
                continue
            data_type = DTYPE_MAP.get(input_object.get(DTYPE_OBJECT))
            if not data_type:
                logger.error(f"The dtype attribute does not support {input_object[DTYPE_OBJECT]} value.")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_KEY_ERROR)
            data_type_size = np.dtype(data_type).itemsize
            if self.shape_range:
                range_shape_size_list = self._get_range_shape_size_list(input_object)
                for item in range_shape_size_list:
                    value.append(item * data_type_size)
            else:
                item_sum = 1
                for num in input_object.get(SHAPE_OBJECT).get(DIM_OBJECT):
                    item_sum *= num
                value.append(item_sum * data_type_size)
        return value

    def _process_inputs_to_list(self, input_desc_array):
        shape_lists = []
        for input_object in input_desc_array:
            shape_list = []
            if SHAPE_OBJECT not in input_object:
                logger.error("Please specify the input shape of om model through -is param")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
            data_type = DTYPE_MAP.get(input_object.get(DTYPE_OBJECT))
            if not data_type:
                logger.error(f"The dtype attribute does not support {input_object[DTYPE_OBJECT]} value.")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_KEY_ERROR)
            if self.shape_range:
                logger.error("Please specify the input shape of om model through -is param")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
            else:
                input_format = input_object.get(LAYOUT_OBJECT, "NCHW")
                input_format_index = [input_format.index(x) for x in "NCHW"]
                for num in input_object.get(SHAPE_OBJECT).get(DIM_OBJECT):
                    shape_list.append(num)
                shape_list = list(np.array(shape_list)[input_format_index])
                shape_lists.append(shape_list)
        return shape_lists



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import math
from enum import Enum
from dataclasses import dataclass

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.log import logger
from msprobe.core.common.exceptions import MsprobeException


def str2float(percentage_str):
    """
    百分比字符串转换转换为浮点型
    Args:
        percentage_str: '0.00%', '23.4%'
    Returns: float 0.00, 0.234
    """
    try:
        percentage_str = percentage_str.strip('%')
        return float(percentage_str) / 100
    except (ValueError, AttributeError):
        return 0


def is_inf_or_nan(value):
    if isinstance(value, float):
        return math.isnan(value) or math.isinf(value)
    elif isinstance(value, str):
        value = value.strip().lower()
        if value in ['inf', '-inf', 'nan']:
            return True
    return False


def divide_result_df(result_df):
    """
    result_df就是输出csv前的数据，按api粒度划分每行数据

    return: dict
        {
            api_name1: [[data1, data2, ...], [data1, data2, ...]],
            api_name2: [[data1, data2, ...], [data1, data2, ...]],
            ...
        }
    """
    result_lists = result_df.values.tolist()
    try:
        api_name_index = result_df.columns.get_loc(Const.API_ORIGIN_NAME)
    except KeyError as e:
        logger.error_log_with_exp(f"An error occurred during get index from dataframe: {e}",
                                  MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
    result_dict = {}
    for result_list in result_lists:
        if not result_list:
            continue
        api_name = result_list[api_name_index]
        if api_name in result_dict:
            result_dict.get(api_name).append(result_list)
        else:
            result_dict[api_name] = [result_list]
    return result_dict


class ResultLevel(Enum):
    PASS = CompareConst.PASS
    ERROR = CompareConst.ERROR
    WARNING = CompareConst.WARNING

    def __lt__(self, other):
        """支持比较，用于汇总结果 ERROR > WARNING > PASS"""
        order = [self.PASS, self.WARNING, self.ERROR]
        return order.index(self) < order.index(other)


class CompareMode(Enum):
    TENSOR = Const.ALL
    STATISTICS = Const.SUMMARY
    MD5 = Const.MD5


@dataclass
class AlgorithmResult:
    """算法执行结果"""
    status: ResultLevel = ResultLevel.PASS
    message: list = None


class IgnoreInfo(Enum):
    ALL_IGNORE = 'all_ignore'
    INPUT_IGNORE = 'input_ignore'
    NO_IGNORE = 'no_ignore'


def get_data_list_by_ignore_info(api_data, ignore_info: IgnoreInfo):
    if ignore_info == IgnoreInfo.ALL_IGNORE:
        data_lists = []
    elif ignore_info == IgnoreInfo.INPUT_IGNORE:
        data_lists = api_data.output_data
    else:
        data_lists = api_data.input_data + api_data.output_data
    return data_lists



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves file-related function.
"""
import csv
import json
import os
import re
import uuid
import numpy as np

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils import utils
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse import dump_utils

OP_TYPE = 1
OP_NAME = 2
TASK_ID = 3
STREAM_ID = 4
TIMESTAMP = 4
CONTEXT_ID = 6
THREAD_ID = 7


class FileUtils:
    """
    The class for file utils
    """
    CSV_SUFFIX = '.csv'

    @staticmethod
    def read_csv(path: str) -> list:
        """
        Read csv file to list.
        :path: csv file path
        """
        content = []
        path = os.path.realpath(path)  # 标准化文件路径
        if not str(path).endswith(FileUtils.CSV_SUFFIX):
            log.print_warn_log('read csv failed, file path'
                               ' [{}] is invalid'.format(path))
            return content
        utils.check_file_size(path, ConstManager.ONE_HUNDRED_MB)
        try:
            with open(path, 'r') as f:
                csv_handle = csv.reader(f)
                for row in csv_handle:
                    content.append(row)
            return content
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as io_error:
            log.print_open_file_error(path, io_error)
            raise CompareError(CompareError.MSACCUCMP_OPEN_FILE_ERROR) from io_error
        finally:
            pass

    @staticmethod
    def save_file(file_path: str, content: any) -> None:
        """
        save txt format result to file.
        :file_path: file path
        :content: txt content
        """
        if os.path.islink(file_path):
            os.unlink(file_path)
        try:
            with os.fdopen(os.open(file_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                           'w+') as output_file:
                output_file.write(content)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            log.print_error_log('Failed to open "%r". %s ' % (file_path, str(error)))
            raise CompareError(CompareError.MSACCUCMP_WRITE_FILE_ERROR) from error
        finally:
            pass

    @staticmethod
    def delete_file(path: str) -> None:
        '''Delete file if it exists and user has permission'''
        try:
            if os.path.exists(path):
                os.remove(path)
        except OSError as error:
            raise CompareError(CompareError.MSACCUCMP_DELETE_FILE_ERROR) from error

    @staticmethod
    def save_data_to_file(path: str, data: any, flag: str, delete: bool) -> None:
        """
        Save data to file.
        :param path: the saved file path
        :param data: the data to save
        :param flag: the write flag
        :param delete: delete the path or not
        """
        if delete:
            FileUtils.delete_file(path)

        if flag == "r":
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

        try:
            with os.fdopen(os.open(path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES), flag) as output_file:
                output_file.write(data)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            log.print_error_log('Failed to write data to "%r". %s ' % (path, str(error)))
            raise CompareError(CompareError.MSACCUCMP_WRITE_FILE_ERROR) from error
        finally:
            pass

    @staticmethod
    def save_array_to_file(path: str, array: np.ndarray, np_save: bool, shape: any = None) -> None:
        """
        Save numpy array to file.
        :param path: the saved file path
        :param array: the numpy array
        :param np_save: save or not
        :param shape: the array shape
        """
        try:
            FileUtils._save_array_to_file(path, array, np_save, shape)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            log.print_error_log('Failed to write data to "%r". %s' % (path, str(error)))
        finally:
            pass

    @staticmethod
    def handle_too_long_file_name(old_file_name: str, suffix: str, mapping_file: str) -> str:
        """
        Handle too log file name and save {new,old} map to file
        :param old_file_name: the old file name
        :param suffix: the file suffix
        :param mapping_file: the mapping file
        """
        if len(old_file_name) >= ConstManager.LINUX_FILE_NAME_MAX_LEN:
            value = ''.join(str(uuid.uuid3(uuid.NAMESPACE_DNS, old_file_name)).split('-'))
            new_file_name = "%s%s" % (value, suffix)
            FileUtils.save_data_to_file(mapping_file, "%s,%s\n" % (new_file_name, old_file_name), 'a+', delete=False)
            return new_file_name
        return old_file_name

    @staticmethod
    def _save_array_to_file(path: str, array: np.ndarray, np_save: bool, shape: any = None) -> None:
        '''
        Save numpy array to file.
        :param path: the saved file path
        :param array: the numpy array
        :param np_save: save or not
        :param shape: the array shape
        '''
        if shape:
            array = array.reshape(shape)

        if os.path.islink(path):
            os.unlink(path)
        if os.path.exists(path):
            os.remove(path)

        with os.fdopen(os.open(path, os.O_WRONLY | os.O_CREAT, ConstManager.WRITE_MODES), 'wb') as f:
            if np_save:
                np.save(f, array)
            else:
                array.tofile(f)
        os.chmod(path, ConstManager.WRITE_MODES)

    @classmethod
    def load_json_file(cls: any, json_file: str) -> any:
        """
        load json file
        :json_file: the json file
        """
        utils.check_file_size(json_file, ConstManager.ONE_HUNDRED_MB)
        try:
            with open(json_file, 'r') as input_file:
                return json.load(input_file)
        except Exception as error:
            log.print_error_log(
                'Failed to load json object. The content of the json file "%r" is invalid.' % json_file)
            raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR) from error


class OverflowFileUtils(FileUtils):
    """
    The class for the overflow files
    """
    # file name patterns
    # example:
    # dump_file: Add.partition0_rank1_new_sub_graph1_sgt_graph_0_Add.0.3.1673383728093446.4.1.0.0
    # opdeug_file: Opdebug.Node_OpDebug.0.3.1673383728094391.4.1.0.0
    # parsed_debug_file: Opdebug.Node_OpDebug.0.3.1673383728094391.4.1.0.0.output.0.json
    DUMP_FILE_PATTERN = r"^([A-Za-z0-9_-]+)\.([A-Za-z0-9_-]+)" \
                        r"\.([0-9]+)\.?([0-9]+)?\.([0-9]{1,255})" \
                        r"\.?([0-9]+)?\.?([0-9]+)?\.?([0-9]+)?\.?([0-9]+)?"
    PARSED_DUMP_FILE_PATTERN = \
        r"^([A-Za-z0-9_-]+)\.([A-Za-z0-9_-]+)\.([0-9]+)\.?([0-9]+)?" \
        r"\.([0-9]{1,255})\.([a-z]+)\.([0-9]{1,255})\.npy$"
    DEBUG_FILE_PATTERN = r"(Opdebug)\.(Node_OpDebug)\.([0-9]+)\.?([0-9]+)?\.([0-9]{1,255})" \
                         r"\.?([0-9]+)?\.?([0-9]+)?\.?([0-9]+)?\.?([0-9]+)?"
    PARSED_DEBUG_FILE_PATTERN = r"Opdebug\.Node_OpDebug\.([0-9]+)\.?([0-9]+)?" \
                                r"\.([0-9]{1,255})(\.[0-9])?\.?([0-9]+)?\.?([0-9]+)?" \
                                r"\.?([0-9]+)?\.([a-z]+)\.([0-9]{1,255})\.json"
    MAX_DEPTH = 30

    @staticmethod
    def _list_file_with_pattern(path: str, pattern: str,
                                extern_pattern: any, gen_info_func: any) -> dict:
        """
        find all files for a specific pattern
        :path: the path of the dir
        :pattern: the pattern
        :extern_pattern: the extern_pattern
        :gen_info_func: a function to gen the file obj
        """
        if not path or not os.path.exists(path):
            raise CompareError(
                CompareError.MSACCUCMP_INVALID_PATH_ERROR)

        if len(path) >= ConstManager.LINUX_PATH_MAX_LEN:
            log.print_error_log(
                "The path length exceeds the maximum limit, "
                f"the maximum limit is {ConstManager.LINUX_PATH_MAX_LEN}.")
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)

        matched_files = {}
        re_pattern = re.compile(pattern)
        for dir_path, _, files in os.walk(path):
            current_depth = dir_path.count(os.sep) - path.count(os.sep)
            if current_depth > OverflowFileUtils.MAX_DEPTH:
                log.print_warn_log(
                    "The current file depth has exceeded the maximum depth limit, "
                    f"which is set to {OverflowFileUtils.MAX_DEPTH}.")
                break
            for file in files:
                match = re_pattern.match(file)
                if match is None:
                    continue
                if extern_pattern != '' and not re.match(extern_pattern, file):
                    continue
                if gen_info_func:
                    matched_files[file] = gen_info_func(file, dir_path, match)
        return matched_files

    @staticmethod
    def _gen_dump_file_info(name: str, dir_path: str, match: any) -> any:
        """
        generate the dump file object
        :name: file name
        :dir_path: file path
        :match: the result of re.match() match with file name
        """

        if len(match.groups()) > 4:
            file_desc = {
                "file_path": os.path.join(dir_path, name),
                "timestamp": int(match.groups()[TIMESTAMP])
            }
            dump_attr = {
                "op_name": match.group(OP_NAME),
                "op_type": match.group(OP_TYPE),
                "task_id": int(match.group(TASK_ID)),
                "stream_id": match.group(STREAM_ID)
            }
        if len(match.groups()) > 7 and match.groups()[CONTEXT_ID] and match.groups()[THREAD_ID]:
            dump_attr["context_id"] = int(match.groups()[CONTEXT_ID])
            dump_attr["thread_id"] = int(match.groups()[THREAD_ID])

        return DumpFileDesc(file_desc, dump_attr)

    @staticmethod
    def _gen_one_tensor_desc(output_path: str, dump_file_desc: any,
                             tensor_list: any, tensor_type: str = '') -> dict:
        """
        generate the ParsedDumpFileDesc of the tensor list
        :output_path: the output path
        :dump_file_desc: the DumpFileDesc for the debug file
        :tensor_list: tensor list of one dump file
        :tensor_type: tensor type, such as "input"/"output"
        """
        if len(tensor_list) == 0:
            return {}

        dump_file_name = os.path.basename(dump_file_desc.file_path)
        parsed_dump_files = {}
        for (index, tensor) in enumerate(tensor_list):
            parsed_dump_file_name = "{}.{}.{}.npy" \
                .format(dump_file_name, tensor_type, index)
            file_desc = {
                "file_path": os.path.join(output_path, parsed_dump_file_name),
                "timestamp": dump_file_desc.timestamp
            }
            dump_attr = {
                "op_name": dump_file_desc.op_name,
                "op_type": dump_file_desc.op_type,
                "task_id": dump_file_desc.task_id,
                "stream_id": dump_file_desc.stream_id
            }
            anchor = {
                "anchor_type": tensor_type,
                "anchor_idx": index,
                "format": common.get_format_string(tensor.tensor_format)
            }

            parsed_dump_files[parsed_dump_file_name] \
                = ParsedDumpFileDesc(file_desc, dump_attr, anchor)
        return parsed_dump_files

    @staticmethod
    def _gen_parsed_debug_file_info(name: str, dir_path: str, match: any) -> object:
        """
        generate the objects of the parsed debug_file.
        :name: parsed file name
        :dir_path: the path of parsed debug file
        :match: the result of re.match() match with \
               the name of parsed debug file
        """
        if len(match.groups()) < 3:
            log.print_error_log(f"Invalid debug file name {name}")
            raise CompareError(CompareError.MSACCUCMP_INVALID_FILE_ERROR)
        file_desc = {
            "file_path": os.path.join(dir_path, name),
            "timestamp": int(match.groups()[2])
        }
        dump_attr = {
            "op_name": 'Node_OpDebug',
            "op_type": 'Opdebug',
            "task_id": int(match.group(1)),
            "stream_id": match.group(2)
        }
        if len(match.groups()) > 5 and match.groups()[4] and match.groups()[5]:
            dump_attr["context_id"] = int(match.groups()[4])
            dump_attr["thread_id"] = int(match.groups()[5])
        anchor = {
            "anchor_type": match.groups()[-2],
            "anchor_idx": int(match.groups()[-1])
        }
        return ParsedDumpFileDesc(file_desc, dump_attr, anchor)

    def parse_mapping_csv(self: any, path: str, pattern: str, extern_pattern: any = '') -> dict:
        """
        parse files in mapping.csv
        :path: the path of the dir
        :pattern: the pattern
        """
        hash_index = 0
        file_name_index = 1
        matched_files = {}
        re_pattern = re.compile(pattern)
        if ConstManager.MAPPING_FILE_NAME not in os.listdir(path):
            return matched_files
        mapping = self.read_csv(os.path.join(path, ConstManager.MAPPING_FILE_NAME))
        for item in mapping:
            src_file = os.path.realpath(os.path.join(path, item[hash_index]))
            if not os.path.isfile(src_file):
                log.print_warn_log("the file %r in mapping.csv is not exist, dir: %r."
                                   % (item[hash_index], path))
                continue
            match = re_pattern.match(item[file_name_index])
            if match is None:
                continue
            if extern_pattern != '' and not re.match(extern_pattern,
                                                     item[file_name_index]):
                continue

            matched_files[item[hash_index]] = self._gen_dump_file_info(
                item[hash_index], path, match)
        return matched_files

    def list_dump_files(self: any, dump_path: str, pattern: str, extern_pattern: any = '') -> any:
        """
        parse all the dump files
        :dump_path: the dump file path
        :pattern: the special matching mode of dump files
        :extern_pattern: the extern_pattern
        """
        npu_dump_files = self._list_file_with_pattern(
            dump_path, pattern, extern_pattern, self._gen_dump_file_info)
        npu_dump_files.update(
            self.parse_mapping_csv(dump_path, pattern, extern_pattern))
        return list(npu_dump_files.values())

    def list_parsed_dump_files(self: any, output_path: str, dump_file_desc: any) -> dict:
        """
        search all the parsed npu dump files
        :output_path: the output path
        :dump_file_desc: the DumpFileDesc for the debug file
        """
        # only support the default dump version ‘2’
        dump_data = dump_utils.parse_dump_file(dump_file_desc.file_path, ConstManager.BINARY_DUMP_TYPE)
        dump_file_desc.set_op_name(dump_data.op_name)
        parsed_dump_files = self._gen_one_tensor_desc(output_path, dump_file_desc,
                                                      dump_data.input_data, 'input')
        parsed_dump_files.update(self._gen_one_tensor_desc(output_path, dump_file_desc,
                                                           dump_data.output_data, 'output'))
        parsed_dump_files.update(self._gen_one_tensor_desc(output_path, dump_file_desc,
                                                           dump_data.buffer))
        return parsed_dump_files

    def list_parsed_debug_files(self: any, dir_path: str, extern_pattern: str = '') -> dict:
        """
        search for all parsed debug files in the path.
        :dir_path: the path
        :extern_pattern: the extern_pattern
        """
        return self._list_file_with_pattern(
            dir_path, self.PARSED_DEBUG_FILE_PATTERN, extern_pattern,
            self._gen_parsed_debug_file_info)


class FileDesc:
    """
    The class for file description
    """

    def __init__(self: any, file_desc: dict) -> None:
        self.file_path = file_desc.get("file_path")
        self.timestamp = file_desc.get("timestamp")
        if not self.timestamp:
            self.timestamp = os.path.getmtime(self.file_path)

    def get_file_path(self: any) -> str:
        """
        get file name
        """
        return self.file_path

    def get_file_time(self: any) -> float:
        """
        get file path
        """
        return self.timestamp


class DumpFileDesc(FileDesc):
    """
    The class for file dump file description
    """

    def __init__(self: any, file_desc: dict, dump_attr: dict) -> None:
        super(DumpFileDesc, self).__init__(file_desc)
        self.op_name = dump_attr.get("op_name")
        self.op_type = dump_attr.get("op_type")
        self.task_id = dump_attr.get("task_id")
        self.stream_id = int(dump_attr.setdefault("stream_id", '0'))
        self.context_id = dump_attr.setdefault("context_id", None)
        self.thread_id = dump_attr.setdefault("thread_id", None)

    def set_op_name(self: any, op_name: str) -> None:
        """
        update op name.
        :op_name: the new op_name
        """
        self.op_name = op_name


class ParsedDumpFileDesc(DumpFileDesc):
    """
    The class for file decoded dump file description
    """

    def __init__(self: any, file_desc: dict, dump_attr: dict, anchor: dict) -> None:
        super(ParsedDumpFileDesc, self).__init__(file_desc, dump_attr)
        self.type = anchor.get("anchor_type")
        self.idx = anchor.get("anchor_idx")
        self.format = anchor.get("format")


class UmaskWrapper:
    """Write with preset umask
    >>> with UmaskWrapper():
    >>>     ...
    """

    def __init__(self, umask=0o027):
        self.umask, self.ori_umask = umask, None

    def __enter__(self):
        self.ori_umask = os.umask(self.umask)

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        os.umask(self.ori_umask)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, List, Optional

from msprobe.core.common.log import logger


class BaseMonitorV2(ABC):
    """
    Minimal base class for v2 monitors.

    Compared to core.monitor.BaseMonitor, this class is intentionally kept
    small so that higher-level policies live in dedicated components
    (framework, trainer, backends) instead of being mixed into the base.
    """

    DEFAULT_OPS = ["min", "max", "mean", "norm", "nans"]

    def __init__(self) -> None:
        self.config: Dict[str, Any] = {}
        self._context: Dict[str, Any] = {}
        self._ops: List[str] = []
        # Common per-step rows buffer; subclasses can populate this and
        # rely on the default collect() implementation.
        self._rows: List[Dict[str, Any]] = []

    def set_config(self, config: Dict[str, Any]) -> None:
        self.config = config
        self._configure_ops(config.get("ops"))

    def set_context(self, **context: Any) -> None:
        self._context.update(context)

    def collect(self) -> Optional[Dict[str, Any]]:
        """
        Default collect implementation used by most monitors:
        return per-step rows and reset the internal buffer.
        """
        if not self._rows:
            return None
        data = {"rows": self._rows}
        self._rows = []
        return data

    @abstractmethod
    def start(self, *args: Any, **kwargs: Any) -> None:
        ...

    @abstractmethod
    def stop(self) -> None:
        ...

    def _configure_ops(self, ops: Optional[Iterable[str]] = None) -> None:
        if ops is None or not isinstance(ops, Iterable):
            self._ops = list(self.DEFAULT_OPS)
            return
        requested = [ops] if isinstance(ops, str) else list(ops)
        invalid_ops = [op for op in requested if op not in self.DEFAULT_OPS]
        if invalid_ops:
            logger.warning(
                f"[monitor_v2] Unsupported ops {invalid_ops}; supported: {self.DEFAULT_OPS}."
            )
        filtered = [op for op in requested if op in self.DEFAULT_OPS]
        if not filtered:
            logger.warning("[monitor_v2] No valid ops provided; fallback to default ops.")
        self._ops = filtered if filtered else list(self.DEFAULT_OPS)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch

from msprobe.core.common.const import CompareConst


class StandardConfig:
    """
    Standard configuration class for managing precision and comparison thresholds.

    This class provides a centralized way to manage the small value thresholds, absolute tolerances,
    and relative tolerances (rtol) used in precision comparisons. It allows for different thresholds
    based on the data type, with default values provided for common data types.

    Attributes:
        _small_value (dict): A dictionary mapping data types to their corresponding small value thresholds.
        _small_value_atol (dict): A dictionary mapping data types to their corresponding absolute tolerances.
        _rtol (dict): A dictionary mapping data types to their corresponding relative tolerances.

    Methods:
        get_small_value(dtype): Retrieves the small value threshold for the given data type.
        get_small_value_atol(dtype): Retrieves the absolute tolerance for the given data type.
        get_rtol(dtype): Retrieves the relative tolerance for the given data type.

    Example:
        >>> small_value = StandardConfig.get_small_value(torch.float32)
        >>> atol = StandardConfig.get_small_value_atol(torch.float32)
        >>> rtol = StandardConfig.get_rtol(torch.float32)
        >>> print(small_value, atol, rtol)
        1e-6 1e-9 1e-6

    Note:
        The data type is expected to be a PyTorch data type. If the data type is not found in the dictionary,
        the default value is returned.

    See Also:
        torch.dtype: PyTorch data types.
    """
    # 维护三份列表（示例）
    L1_API_LIST = { 
        "embedding", "npu_fusion_attention_grad",
        "npu_fusion_attention", "scatter_", "scatter_add_"
    }
    L2_API_LIST = {
        "addmm", "npu_cross_entropy_loss", "npu_cross_entropy_loss_backward"
        ,"GELU", "layer_norm", "matmul", "mm", "index", "one_hot", "sum",
        "npu_rms_norm", "npu_rms_norm_backward", "nonzero"
    } 
    _small_value = {
        torch.float16: 2**-10,
        torch.bfloat16: 2**-10,
        torch.float32: 2**-20,
        "default": 2**-20
    }
    _threshold_small_value_atol = {
        torch.float16: 2**-16,
        torch.bfloat16: 1e-16,
        torch.float32: 2**-30,
        "default": 2**-30
    }
    _benchmark_small_value_atol = {
        torch.float16: 1e-16,
        torch.bfloat16: 1e-16,
        torch.float32: 2**-30,
        "default": 2**-30
    }
    _rtol = {
        torch.float16: 2**-10,
        torch.bfloat16: 2**-8,
        torch.float32: 2**-20,
        "default": 2**-20
    }
    _accumulative_error_bound = {
        torch.float16: 2**-8,
        torch.bfloat16: 2**-7,
        torch.float32: 2**-11,
        "default": 2**-11
    }
    _small_value_threshold = {
        'error_threshold': 2,
        'warning_threshold': 1,
        "default": 1
    }
    _rmse_threshold = {
        'error_threshold': 2,
        'warning_threshold': 1,
        "default": 1
    }
    _max_rel_err_threshold = {
        'error_threshold': 10,
        'warning_threshold': 1,
        "default": 1
    }
    _mean_rel_err_threshold = {
        'error_threshold': 2,
        'warning_threshold': 1,
        "default": 1
    }
    _eb_threshold = {
        'error_threshold': 2,
        'warning_threshold': 1,
        "default": 1
    }
    _minmum_err = {
        'torch.float16': 2**-11,
        'torch.bfloat16': 2**-8,
        'torch.float32': 2**-14,
        'default': 2**-14
    }
    _accumulative_error_eb_threshold = {
        'torch.float16': 2**-20,
        'torch.bfloat16': 2**-7,
        'torch.float32': 2**-14,
        'default': 2**-14
    }
    
    _fp32_mean_ulp_err_threshold = 64
    ulp_err_proportion_ratio = 1
    _fp32_ulp_err_proportion = 0.05
    _fp16_ulp_err_proportion = 0.001
    _special_samll_value = 1
    
    @classmethod
    def get_api_level(cls, api_name: str) -> str:
        if api_name in cls.L2_API_LIST:
            return "L2"
        if api_name in cls.L1_API_LIST:
            return "L1"
        return "L0"

    @classmethod
    def get_benchmark_ratio_threshold(cls, metric, api_name: str = None, level: str = None):
        # 三档阈值
        thresholds = {
            "L0": {
                CompareConst.MAX_REL_ERR: 10.0,
                CompareConst.MEAN_REL_ERR: 2.0,
                CompareConst.RMSE: 2.0,
            },
            "L1": {
                CompareConst.MAX_REL_ERR: 5.0,
                CompareConst.MEAN_REL_ERR: 1.5,
                CompareConst.RMSE: 1.5,
            },
            "L2": {
                CompareConst.MAX_REL_ERR: 2.0,
                CompareConst.MEAN_REL_ERR: 1.2,
                CompareConst.RMSE: 1.2,
            }
        }

        if level is None:
            level = cls.get_api_level(api_name) if api_name else "L0"

        # 其他指标不参与判定：inf
        if metric in (CompareConst.MAX_REL_ERR ,CompareConst.MEAN_REL_ERR ,CompareConst.RMSE):
            return thresholds.get(level, thresholds["L0"]).get(metric, float("inf"))
        else:
            return StandardConfig.get_benchmark_threshold(metric)

    @classmethod
    def get_small_value(cls, dtype, standard):
        if standard == CompareConst.ACCUMULATIVE_ERROR_COMPARE:
            return cls._special_samll_value
        return cls._small_value.get(dtype, cls._small_value["default"])
    
    @classmethod
    def get_small_value_atol(cls, dtype, standard):
        standard_dict = {
            CompareConst.ABSOLUTE_THRESHOLD: cls._threshold_small_value_atol,
            CompareConst.BENCHMARK: cls._benchmark_small_value_atol
        }
        small_value_atol_standard = standard_dict.get(standard, cls._benchmark_small_value_atol)
        return small_value_atol_standard.get(dtype, small_value_atol_standard["default"])
    
    @classmethod
    def get_rtol(cls, dtype):
        return cls._rtol.get(dtype, cls._rtol["default"])
    
    @classmethod
    def get_small_value_threshold(cls, threshold_type):
        return cls._small_value_threshold.get(threshold_type, cls._small_value_threshold["default"])
    
    @classmethod
    def get_rmse_threshold(cls, threshold_type):
        return cls._rmse_threshold.get(threshold_type, cls._rmse_threshold["default"])
    
    @classmethod
    def get_max_rel_err_threshold(cls, threshold_type):
        return cls._max_rel_err_threshold.get(threshold_type, cls._max_rel_err_threshold["default"])
    
    @classmethod
    def get_mean_rel_err_threshold(cls, threshold_type):
        return cls._mean_rel_err_threshold.get(threshold_type, cls._mean_rel_err_threshold["default"])
    
    @classmethod
    def get_eb_threshold(cls, threshold_type):
        return cls._eb_threshold.get(threshold_type, cls._eb_threshold["default"])
    
    @classmethod
    def get_benchmark_threshold(cls, metric):
        metric_threshold_functions = {
            'small_value': StandardConfig.get_small_value_threshold,
            'rmse': StandardConfig.get_rmse_threshold,
            'max_rel_err': StandardConfig.get_max_rel_err_threshold,
            'mean_rel_err': StandardConfig.get_mean_rel_err_threshold,
            'eb': StandardConfig.get_eb_threshold
        }
        
        threshold_func = metric_threshold_functions.get(metric)
        return threshold_func('error_threshold')

    @classmethod
    def get_fp32_mean_ulp_err_threshold(cls):
        return cls._fp32_mean_ulp_err_threshold

    @classmethod
    def get_ulp_err_proportion_ratio_threshold(cls):
        return cls.ulp_err_proportion_ratio

    @classmethod
    def get_fp32_ulp_err_proportion_threshold(cls):
        return cls._fp32_ulp_err_proportion

    @classmethod
    def get_fp16_ulp_err_proportion_threshold(cls):
        return cls._fp16_ulp_err_proportion
    
    @classmethod
    def get_ulp_threshold(cls, dtype):
        ulp_err_proportion_ratio_threshold = StandardConfig.get_ulp_err_proportion_ratio_threshold()
        if dtype == torch.float32:
            mean_ulp_err_threshold = StandardConfig.get_fp32_mean_ulp_err_threshold()
            ulp_err_proportion_threshold = StandardConfig.get_fp32_ulp_err_proportion_threshold()
            return mean_ulp_err_threshold, ulp_err_proportion_threshold, ulp_err_proportion_ratio_threshold
        else:
            ulp_err_proportion_threshold = StandardConfig.get_fp16_ulp_err_proportion_threshold()
            return None, ulp_err_proportion_threshold, ulp_err_proportion_ratio_threshold

    @classmethod
    def get_minmum_err(cls, dtype):
        return cls._minmum_err.get(dtype, cls._minmum_err["default"])
    
    @classmethod
    def get_accumulative_error_bound(cls, dtype):
        return cls._accumulative_error_bound.get(dtype, cls._accumulative_error_bound["default"])
    
    @classmethod
    def get_accumulative_error_eb_threshold(cls, dtype):
        return cls._accumulative_error_eb_threshold.get(dtype, cls._accumulative_error_eb_threshold["default"])



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the nano dump function.
"""
import os
import struct
import warnings
from typing.io import BinaryIO
from enum import Enum

from msprobe.msaccucmp.cmp_utils import path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.tlv_parse import TLV


NANO_DUMP_DATA_MAGIC_NUM = 0x5a5a5a5a
TLV_TYPE_INPUT_DESC = 4
TLV_TYPE_OUTPUT_DESC = 5

INPUT_DESC_TLV_TYPE_SHAPE_DIMS = 0
INPUT_DESC_TLV_TYPE_ORI_SHAPE_DIMS = 1

OUTPUT_DESC_L3_TLV_TYPE_SHAPE_DIMS = 0
OUTPUT_DESC_L3_TLV_TYPE_ORI_SHAPE_DIMS = 1
OUTPUT_DESC_L3_TLV_TYPE_ORI_NAME = 2

INPUT_INFO_CONFIG_TAB = [
    {'Name': 'data_type', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'format', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'address_type', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'address', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT64'},
    {'Name': 'offset', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT64'},
    {'Name': 'size', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT64'},
    {'Name': 'tlv_list_len', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT32'},
    {'Name': 'shape_dims', 'TLV_Type': 'TLNV', 'Ele_Type': 'UINT64',
     'Tag': INPUT_DESC_TLV_TYPE_SHAPE_DIMS
     },
    {'Name': 'original_shape_dims', 'TLV_Type': 'TLNV', 'Ele_Type': 'UINT64',
     'Tag': INPUT_DESC_TLV_TYPE_ORI_SHAPE_DIMS
     }
]

INPUTS_LIST_CONFIG_TAB = [
    {'Name': 'inputs_num', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT32'},
    {'Name': 'inputs', 'TLV_Type': 'NV', 'Ele_Type': INPUT_INFO_CONFIG_TAB, 'N': 'inputs_num'}
]

INPUTS_CONFIG_TAB = [
    {'Name': 'input_desc', 'TLV_Type': 'TLV', 'Ele_Type': INPUTS_LIST_CONFIG_TAB, 'Tag': TLV_TYPE_INPUT_DESC}
]

OUTPUT_INFO_CONFIG_TAB = [
    {'Name': 'data_type', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'format', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'address_type', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'original_index', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'original_data_type', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'original_format', 'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'address', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT64'},
    {'Name': 'offset', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT64'},
    {'Name': 'size', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT64'},
    {'Name': 'tlv_list_len', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT32'},
    {'Name': 'shape_dims', 'TLV_Type': 'TLNV', 'Ele_Type': 'UINT64',
     'Tag': OUTPUT_DESC_L3_TLV_TYPE_SHAPE_DIMS
     },
    {'Name': 'original_shape_dims', 'TLV_Type': 'TLNV', 'Ele_Type': 'UINT64',
     'Tag': OUTPUT_DESC_L3_TLV_TYPE_ORI_SHAPE_DIMS
     },
    {'Name': 'origin_name', 'TLV_Type': 'TLNV', 'Ele_Type': 'CHAR',
     'Tag': OUTPUT_DESC_L3_TLV_TYPE_ORI_NAME
     }
]

OUTPUTS_LIST_CONFIG_TAB = [
    {'Name': 'outputs_num', 'TLV_Type': 'ATOM', 'Ele_Type': 'UINT32'},
    {'Name': 'outputs', 'TLV_Type': 'NV', 'Ele_Type': OUTPUT_INFO_CONFIG_TAB, 'N': 'outputs_num'}
]

OUTPUTS_CONFIG_TAB = [
    {'Name': 'output_desc', 'TLV_Type': 'TLV', 'Ele_Type': OUTPUTS_LIST_CONFIG_TAB, 'Tag': TLV_TYPE_OUTPUT_DESC}
]


class NanoDataType(Enum):
    DT_FLOAT = 0
    DT_FLOAT16 = 1
    DT_INT8 = 2
    DT_INT16 = 6
    DT_UINT16 = 7
    DT_UINT8 = 4
    DT_INT32 = 3
    DT_INT64 = 9
    DT_UINT32 = 8
    DT_UINT64 = 10
    DT_BOOL = 12
    DT_DOUBLE = 11
    DT_STRING = 13
    DT_DUAL_SUB_INT8 = 14
    DT_DUAL_SUB_UINT8 = 15
    DT_COMPLEX64 = 16
    DT_COMPLEX128 = 17
    DT_QINT8 = 18
    DT_QINT16 = 19
    DT_QINT32 = 20
    DT_QUINT8 = 21
    DT_QUINT16 = 22,
    DT_RESOURCE = 23
    DT_STRING_REF = 24
    DT_DUAL = 25
    DT_VARIANT = 26
    DT_BF16 = 27
    DT_UNDEFINED = 28
    DT_INT4 = 29
    DT_UINT1 = 30
    DT_INT2 = 31
    DT_UINT2 = 32
    DT_MAX = 33


NANO_DATA_TYPE_TO_PROTO_DATA_TYPE = {
    NanoDataType.DT_FLOAT.value: DD.DT_FLOAT,
    NanoDataType.DT_FLOAT16.value: DD.DT_FLOAT16,
    NanoDataType.DT_DOUBLE.value: DD.DT_DOUBLE,
    NanoDataType.DT_INT8.value: DD.DT_INT8,
    NanoDataType.DT_INT16.value: DD.DT_INT16,
    NanoDataType.DT_INT32.value: DD.DT_INT32,
    NanoDataType.DT_INT64.value: DD.DT_INT64,
    NanoDataType.DT_UINT8.value: DD.DT_UINT8,
    NanoDataType.DT_UINT16.value: DD.DT_UINT16,
    NanoDataType.DT_UINT32.value: DD.DT_UINT32,
    NanoDataType.DT_UINT64.value: DD.DT_UINT64,
    NanoDataType.DT_BOOL.value: DD.DT_BOOL,
    NanoDataType.DT_COMPLEX64.value: DD.DT_COMPLEX64,
    NanoDataType.DT_COMPLEX128.value: DD.DT_COMPLEX128,
    NanoDataType.DT_BF16.value: DD.DT_BF16,
    NanoDataType.DT_UINT1.value: DD.DT_UINT1,
    NanoDataType.DT_RESOURCE.value: DD.DT_INT64,
}


class EmptyObj:
    pass


class NanoDumpData:
    def __init__(self, op_type="", op_name="", dump_time=0) -> None:
        self.magic_num = None
        self.version_id = 0
        self.op_type = op_type
        self.op_name = op_name
        self.dump_time = dump_time
        self.inputs = []
        self.outputs = []

    @staticmethod
    def _unpack_value(aux: bytes, value_type: str) -> (bytes, any):
        type_para = ConstManager.UNPACK_FORMAT.get(value_type)
        _fmt, _size = type_para.get('FMT'), type_para.get('SIZE')

        if len(aux) < _size:
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)

        value = struct.unpack(_fmt, aux[:_size])[0]
        return aux[_size:], value

    def parse_tlv_head(self, content: bytes) -> None:
        aux = content

        # magic_num
        aux, self.magic_num = self._parse_uint32(aux)
        if self.magic_num != NANO_DUMP_DATA_MAGIC_NUM:
            # parse data from bin file stream
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)

        # version_id
        aux, self.version_id = self._parse_uint32(aux)

        # inputs
        intputs_desc_obj = EmptyObj()
        tlv = TLV(INPUTS_CONFIG_TAB)
        try:
            aux, nested_obj = tlv.parse_tlv_by_cfg_tab(aux, intputs_desc_obj)
        except CompareError as error:
            log.print_error_log('Failed to decode inputs info %s', error)
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from error

        if nested_obj.input_desc is not None:
            self.inputs = nested_obj.input_desc.inputs

        # outputs
        outputs_desc_obj = EmptyObj()
        tlv = TLV(OUTPUTS_CONFIG_TAB)
        try:
            aux, nested_obj = tlv.parse_tlv_by_cfg_tab(aux, outputs_desc_obj)
        except CompareError as error:
            log.print_error_log('Failed to decode outputs info %s' % error)
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from error

        if nested_obj.output_desc is not None:
            self.outputs = nested_obj.output_desc.outputs

        self._convert_data_type_to_proto_data_type()

    def _parse_uint32(self, aux: bytes) -> (bytes, any):
        return self._unpack_value(aux, 'UINT32')

    def _convert_data_type_to_proto_data_type(self):
        for index, item in enumerate(self.inputs):
            self.inputs[index].data_type = NANO_DATA_TYPE_TO_PROTO_DATA_TYPE.get(item.data_type)

        for index, item in enumerate(self.outputs):
            self.outputs[index].data_type = NANO_DATA_TYPE_TO_PROTO_DATA_TYPE.get(item.data_type)
            self.outputs[index].original_data_type = NANO_DATA_TYPE_TO_PROTO_DATA_TYPE.get(item.original_data_type)


class NanoDumpDataParser:
    """
    The class for big dump data parser
    """
    warnings.filterwarnings("ignore")

    def __init__(self: any, dump_file_path: str) -> None:
        self.dump_file_path = dump_file_path
        self.file_size = os.path.getsize(self.dump_file_path)
        self.tlv_header_length = 0

        file_name = os.path.basename(self.dump_file_path)
        op_name = file_name.split('.')[1]
        dump_time = file_name.split('.')[4]
        self.nano_dump_data = NanoDumpData(op_name, dump_time)

    def parse(self: any) -> NanoDumpData:
        """
        Parse the dump file path by nano dump data format

        file format:
        |tlv head length| tlv head |input0 data |input1 data |...|output0 data |output1 data |...|
        |uint64         | tlv head |input0 data |input1 data |...|output0 data |output1 data |...|

        tlv head format:
        |——uint32 magicnum  //0x5A5A5A5A
        |——uint32 version_id
        |——list of tensor(input和output)

        :return: NanoDumpData
        :exception when read or parse file error
        """
        with open(self.dump_file_path, 'rb') as dump_file:
            self._read_header_length(dump_file)
            self._read_tlv_head_data(dump_file)
            self._check_size_match()
            self._read_input_data(dump_file)
            self._read_output_data(dump_file)
            return self.nano_dump_data

    def _check_size_match(self: any) -> None:
        input_data_size = 0
        for item in self.nano_dump_data.inputs:
            input_data_size += item.size
        output_data_size = 0
        for item in self.nano_dump_data.outputs:
            output_data_size += item.size

        if self.tlv_header_length + input_data_size + output_data_size + \
                ConstManager.UINT64_SIZE != self.file_size:
            log.print_error_log(
                'The file size (%d) of %r is not equal to %d (tlv_header_length)'
                '+ %d(the sum of input data) + %d(the sum of output data) '
                '. Please check the dump file.'
                % (self.file_size, self.dump_file_path,
                   self.tlv_header_length, input_data_size, output_data_size))
            raise CompareError(CompareError.MSACCUCMP_UNMATCH_STANDARD_DUMP_SIZE)

    def _read_header_length(self: any, dump_file: BinaryIO) -> None:
        # read tlv header length
        tlv_header_length = dump_file.read(ConstManager.UINT64_SIZE)
        self.tlv_header_length = struct.unpack(ConstManager.UINT64_FMT, tlv_header_length)[0]

        if self.tlv_header_length > self.file_size - ConstManager.UINT64_SIZE:
            log.print_warn_log(
                'The header content size (%d) of %r must be less than or'
                ' equal to %d (file size) - %d (tlv header length).'
                ' Please check the dump file.'
                % (self.tlv_header_length, self.dump_file_path, self.file_size, ConstManager.UINT64_SIZE))
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)

    def _read_tlv_head_data(self: any, dump_file: BinaryIO) -> None:
        content = dump_file.read(self.tlv_header_length)
        try:
            self.nano_dump_data.parse_tlv_head(content)
        except CompareError as de_error:
            log.print_warn_log(
                'Failed to parse the serialized header content of %r. '
                'Please check the dump file. %s '
                % (self.dump_file_path, str(de_error)))
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from de_error

    def _read_input_data(self: any, dump_file: BinaryIO) -> None:
        for data_input in self.nano_dump_data.inputs:
            setattr(data_input, 'data', dump_file.read(data_input.size))

    def _read_output_data(self: any, dump_file: BinaryIO) -> None:
        for data_output in self.nano_dump_data.outputs:
            setattr(data_output, 'data', dump_file.read(data_output.size))


class NanoDumpDataHandler:
    """
    Handle dump data
    """

    def __init__(self: any, dump_file_path: str) -> None:
        self.dump_file_path = dump_file_path
        self.file_size = 0

    def check_is_nano_dump_format(self):
        ret = path_check.check_path_valid(self.dump_file_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            return False

        self.file_size = os.path.getsize(self.dump_file_path)
        head_len = ConstManager.UINT64_SIZE + ConstManager.UINT32_SIZE
        if self.file_size > head_len:
            with open(self.dump_file_path, 'rb') as dump_file:
                # read tlv_head_len
                _ = dump_file.read(ConstManager.UINT64_SIZE)

                # read magicnum
                magicnum = dump_file.read(ConstManager.UINT32_SIZE)
                magicnum = struct.unpack(ConstManager.UINT32_FMT, magicnum)[0]
                return True if magicnum == NANO_DUMP_DATA_MAGIC_NUM else False
        return False

    def check_argument_valid(self: any) -> None:
        """
        check argument valid
        :exception when invalid
        """
        file_name = os.path.basename(self.dump_file_path)

        if len(file_name.split('.')) < 5:
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, "The file name has no dump time")

        if not self.check_is_nano_dump_format():
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)

        head_len = ConstManager.UINT64_SIZE * 2 + ConstManager.UINT32_SIZE * 2
        if self.file_size < head_len:
            log.print_error_log(
                'The file size (%d) of %r must to bigger than%d (head_len).'
                'Please check the dump file.'
                % (self.file_size, self.dump_file_path, head_len))

        if self.file_size > ConstManager.ONE_GB:
            log.print_warn_log(
                'The size (%d) of %r exceeds 1GB, it may task more time to run, please wait.'
                % (self.file_size, self.dump_file_path))

    def parse_dump_data(self: any) -> NanoDumpData:
        """
        Parse dump file
        :param dump_version: the dump version
        :return: DumpData
        """
        self.check_argument_valid()
        try:
            nano_dump_parser = NanoDumpDataParser(self.dump_file_path)
            return nano_dump_parser.parse()
        except CompareError as error:
            message = 'Failed to parse the dump file %r, type is nano dump format. Please check the dump file. %s' \
                      % (self.dump_file_path, str(error))
            log.print_error_log(message)
            raise error



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


from mindspore import nn
from mindspore import communication
from msprobe.core.common.log import logger
from msprobe.mindspore.common.utils import is_mindtorch
if is_mindtorch():
    import torch


def is_valid_instance(model):
    return isinstance(model, torch.nn.Module) if is_mindtorch() else isinstance(model, nn.Cell)


def get_submodules(model):
    if not is_valid_instance(model):
        logger.info("Counter invalid model, nothing to hook")
        return {}
    return model.named_modules() if is_mindtorch() else model.cells_and_names()


def get_parameters(model):
    if not is_valid_instance(model):
        return {}
    if is_mindtorch():
        return model.named_parameters()
    else:
        return model.parameters_and_names()


def get_rank():
    if comm_is_initialized():
        return communication.get_rank()
    return 0


def comm_is_initialized():
    return communication.GlobalComm.INITED



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import json

import pandas as pd

from msprobe.core.common.file_utils import load_json, load_yaml, create_file_with_content, create_file_in_zip
from msprobe.core.config_check.checkers.base_checker import BaseChecker
from msprobe.core.config_check.config_checker import register_checker_item
from msprobe.core.config_check.utils.utils import config_checking_print, process_pass_check
from msprobe.core.common.const import Const


dirpath = os.path.dirname(__file__)
env_yaml_path = os.path.join(dirpath, "../resource/env.yaml")


def collect_env_data():
    result = {}
    for key, value in os.environ.items():
        result[key] = value
    return result


def get_device_type(env_json):
    for key in env_json.keys():
        if Const.ASCEND in key:
            return Const.NPU_LOWERCASE
    return Const.GPU_LOWERCASE


def compare_env_data(npu_path, bench_path):
    necessary_env = load_yaml(env_yaml_path)
    cmp_data = load_json(npu_path)
    cmp_type = get_device_type(cmp_data)
    bench_data = load_json(bench_path)
    bench_type = get_device_type(bench_data)
    data = []
    for _, value in necessary_env.items():
        cmp_env = value.get(cmp_type)
        bench_env = value.get(bench_type)
        if not bench_env and not cmp_env:
            continue
        elif cmp_env:
            cmp_env_name = cmp_env["name"]
            cmp_value = cmp_data.get(cmp_env_name, value[cmp_type]["default_value"])
            if not bench_env:
                data.append(["only cmp has this env", cmp_env["name"], "", cmp_value, Const.CONFIG_CHECK_WARNING])
                continue
            bench_env_name = bench_env["name"]
            bench_value = bench_data.get(bench_env_name, value[bench_type]["default_value"])
            if cmp_value != bench_value:
                data.append([bench_env_name, cmp_env_name, bench_value, cmp_value, Const.CONFIG_CHECK_ERROR])
        else:
            bench_env_name = bench_env["name"]
            bench_value = bench_data.get(bench_env_name) if bench_data.get(bench_env_name) else value[bench_type][
                "default_value"]
            data.append([bench_env_name, "only bench has this env", bench_value, "", Const.CONFIG_CHECK_WARNING])
    df = pd.DataFrame(data, columns=EnvArgsChecker.result_header)
    return df


@register_checker_item("env")
class EnvArgsChecker(BaseChecker):

    target_name_in_zip = "env"
    result_header = ["bench_env_name", "cmp_env_name", "bench_value", "cmp_value", "level"]

    @staticmethod
    def pack(pack_input):
        output_zip_path = pack_input.output_zip_path
        env_args_dict = collect_env_data()
        create_file_in_zip(output_zip_path, EnvArgsChecker.target_name_in_zip, json.dumps(env_args_dict, indent=4))
        config_checking_print(f"add env args to zip")

    @staticmethod
    def compare(bench_dir, cmp_dir, output_path):
        bench_env_data = os.path.join(bench_dir, EnvArgsChecker.target_name_in_zip)
        cmp_env_data = os.path.join(cmp_dir, EnvArgsChecker.target_name_in_zip)
        df = compare_env_data(bench_env_data, cmp_env_data)
        pass_check = process_pass_check(df['level'].values)
        return EnvArgsChecker.target_name_in_zip, pass_check, df



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import json
import pandas as pd
from msprobe.core.common.file_utils import create_file_in_zip, load_json
from msprobe.core.config_check.checkers.base_checker import BaseChecker
from msprobe.core.config_check.config_checker import register_checker_item, register_pre_forward_fun_list
from msprobe.core.config_check.utils.utils import config_checking_print, get_tensor_features
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.const import Const


@recursion_depth_decorator("config_check: process_obj")
def process_obj(obj):
    if FmkAdp.is_tensor(obj):
        return get_tensor_features(obj)
    elif isinstance(obj, (tuple, list)):
        return {i: process_obj(x) for i, x in enumerate(obj)}
    elif isinstance(obj, dict):
        return {k: process_obj(v) for k, v in obj.items()}
    else:
        return ""


def parse_args_and_kargs(args, kwargs):
    processed_args = process_obj(args)
    processed_kargs = process_obj(kwargs)

    return {
        'args': processed_args,
        'kwargs': processed_kargs
    }


@recursion_depth_decorator("config_check: compare_dataset_dicts")
def compare_dataset_dicts(dict1, dict2, tag=''):
    results = []
    # 处理 dict1 中的键
    for key in dict1:
        new_tag = f"{tag}.{key}" if tag else key
        if key not in dict2:
            result = {'tag': new_tag, 'equal': False, 'status': 'delete'}
            results.append(result)
            continue
        value1 = dict1[key]
        value2 = dict2[key]
        if not isinstance(value1, dict):
            continue
        if set(value1.keys()) == {'max', 'min', 'mean', 'norm'}:
            equal = value1 == value2
            relative_diffs = {
                f"{k}_relative_diff": (abs(value1[k] - value2[k]) / value1[k]) if value1[k] != 0 else None
                for k in ['max', 'min', 'mean', 'norm']
            }
            result = {'tag': new_tag, 'equal': equal, 'status': 'unchanged'}
            result.update(relative_diffs)
            results.append(result)
        else:
            results.extend(compare_dataset_dicts(value1, value2, new_tag))
    # 处理 dict2 中独有的键
    for key in dict2:
        if key not in dict1:
            new_tag = f"{tag}.{key}" if tag else key
            result = {'tag': new_tag, 'equal': False, 'status': 'added'}
            results.append(result)
    return results


def compare_dataset(bench_dir, cmp_dir):
    all_results = []
    for step in os.listdir(bench_dir):
        step_path_bench = os.path.join(bench_dir, step)
        if not os.path.isdir(step_path_bench):
            continue
        step_path_cmp = os.path.join(cmp_dir, step)
        for rank in os.listdir(step_path_bench):
            rank_path_bench = os.path.join(step_path_bench, rank, 'dataset.json')
            rank_path_cmp = os.path.join(step_path_cmp, rank, 'dataset.json')
            if not os.path.isfile(rank_path_bench) or not os.path.isfile(rank_path_cmp):
                continue

            dict1 = load_json(rank_path_bench)
            dict2 = load_json(rank_path_cmp)
            results = compare_dataset_dicts(dict1, dict2)
            for result in results:
                result['step'] = int(step.replace("step", ""))
                result['rank'] = int(rank.replace("rank", ""))
            all_results.extend(results)

    df = pd.DataFrame(all_results, columns=DatasetChecker.result_header)
    df = df.sort_values(by=['step', 'rank'], ascending=[True, True])
    return df


@register_checker_item("dataset")
class DatasetChecker(BaseChecker):
    input_needed = "model"
    multi_rank = True

    target_name_in_zip = "dataset"
    result_header = ['step', 'rank', 'tag', 'equal', 'max_relative_diff',
                     'min_relative_diff', 'mean_relative_diff', 'norm_relative_diff']

    @staticmethod
    def pack(pack_input):
        output_zip_path = pack_input.output_zip_path

        def collect_input(model, args, kwargs, step):
            features = parse_args_and_kargs(args, kwargs)
            dataset_filepath = os.path.join(DatasetChecker.target_name_in_zip, 
                                            f"step{step}", f"rank{FmkAdp.get_rank_id()}", "dataset.json")
            create_file_in_zip(output_zip_path, dataset_filepath, json.dumps(features, indent=4))
            config_checking_print(f"add first dataset input features to zip")
            
        register_pre_forward_fun_list(collect_input)

    @staticmethod
    def compare(bench_dir, cmp_dir, output_path):
        bench_dataset_pack_path = os.path.join(bench_dir, DatasetChecker.target_name_in_zip)
        cmp_dataset_pack_path = os.path.join(cmp_dir, DatasetChecker.target_name_in_zip)

        df = compare_dataset(bench_dataset_pack_path, cmp_dataset_pack_path)
        pass_check = Const.CONFIG_CHECK_PASS if False not in df['equal'].values else Const.CONFIG_CHECK_ERROR
        return DatasetChecker.target_name_in_zip, pass_check, df



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Dict, List, Optional, Any

from msprobe.core.common.const import Const

from msprobe.core.overflow_check.abnormal_scene import InputAnomalyOutputNormalScene, InputAnomalyOutputAnomalyScene, \
    InputNormalOutputAnomalyScene, NumericalMutationScene, AnomalyScene
from msprobe.core.overflow_check.api_info import APIInfo
from msprobe.core.overflow_check.filter import IgnoreFilter
from msprobe.core.overflow_check.level import OverflowLevel


class StatisticsFields:
    """统计字段常量类"""
    CRITICAL_APIS = 'critical_apis'
    HIGH_PRIORITY_APIS = 'high_priority_apis'
    MEDIUM_PRIORITY_APIS = 'medium_priority_apis'
    ANOMALY_DETAILS = 'anomaly_details'

    # 所有字段
    ALL_FIELDS = [CRITICAL_APIS, HIGH_PRIORITY_APIS, MEDIUM_PRIORITY_APIS, ANOMALY_DETAILS]


class AnomalyDetector:
    """异常检测器"""

    def __init__(self, dump_data: Dict):
        """
            初始化检测器，并保存dump_data
        Args:
            dump_data: 数据格式如下
                {
                    "api/module": {statistics}
                }
        """
        self.dump_data = dump_data
        self.ignore_filter = IgnoreFilter()
        self.scene_types = [
            InputNormalOutputAnomalyScene,      # 输入正常，输出异常
            InputAnomalyOutputAnomalyScene,     # 输入异常，输出异常
            InputAnomalyOutputNormalScene,      # 输入异常，输出正常
            NumericalMutationScene              # 输出较输入值突变
        ]
        self.anomaly_scenes: Dict[str, AnomalyScene] = dict()

    @staticmethod
    def _create_api_info(api_name: str, data: Dict) -> APIInfo:
        """从原始数据创建APIInfo实例"""
        return APIInfo(
            api_name=api_name,
            input_args=data.get(Const.INPUT_ARGS, data.get(Const.INPUT, [])),
            input_kwargs=data.get(Const.INPUT_KWARGS, {}),
            output_data=data.get(Const.OUTPUT, [])
        )

    def get_statistics(self) -> Dict[str, List]:
        """获取统计信息

        使用StatisticsFields类统一管理字段名称，避免硬编码

        Returns:
            Dict[str, List]: 包含各优先级API列表和异常详情的字典
        """
        stats = {field: [] for field in StatisticsFields.ALL_FIELDS}

        # 定义rank到结果key的映射关系
        rank_to_key = {
            OverflowLevel.CRITICAL: StatisticsFields.CRITICAL_APIS,
            OverflowLevel.HIGH: StatisticsFields.HIGH_PRIORITY_APIS,
            OverflowLevel.MEDIUM: StatisticsFields.MEDIUM_PRIORITY_APIS
        }

        for scene in self.anomaly_scenes.values():
            stats[StatisticsFields.ANOMALY_DETAILS].append(scene.get_details())
            # 根据rank分类API
            key = rank_to_key.get(scene.rank, None)
            if not key:
                stats[key].append(scene.api_name)

        return stats

    def analyze(self):
        """
            按照异常场景对调用数据进行分析
        Returns:
            返回类本身，若不进行过滤，则仅调用analyze即可
        """
        # 遍历data item
        for api_name, data in self.dump_data.items():
            api_info = self._create_api_info(api_name, data)

            # 每种都进行检测，可能涉及多种命中，原则如下：
            #   - 就高原则
            #   - 优先原则，数据异常放最后检测
            for scene_type in self.scene_types:
                scene = scene_type(api_info)
                if hasattr(scene, 'matches') and scene.matches():
                    self.anomaly_scenes[api_name] = scene
                    break  # 直接跳过，就高原则
        return self

    def filter(self):
        """
            对误检数据进行过滤
        Returns:
            检查checker自身，方便链式调用
        """
        result = dict()
        for api_name, scene in self.anomaly_scenes.items():
            if self.ignore_filter.apply_filter(scene.api_data):
                continue
            result[api_name] = scene
        self.anomaly_scenes = result
        return self

    def overflow_result(self) -> Dict[str, AnomalyScene]:
        return self.anomaly_scenes

    def has_overflow(self, api_name: str) -> bool:
        return api_name in self.anomaly_scenes.keys()

    def get_overflow_level(self, api_name: str) -> Optional[Any]:
        scene = self.anomaly_scenes.get(api_name, None)
        return scene.rank if scene else None
